#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  Crazyflie Python Library
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example script that plots the output ranges from the Multiranger and Flow
deck in a 3D plot.

When the application is started the Crazyflie will hover at 0.3 m. The
Crazyflie can then be controlled by using keyboard input:
 * Move by using the arrow keys (left/right/forward/backwards)
 * Adjust the right with w/s (0.1 m for each keypress)
 * Yaw slowly using a/d (CCW/CW)
 * Yaw fast using z/x (CCW/CW)

There's additional setting for (see constants below):
 * Plotting the downwards sensor
 * Plotting the estimated Crazyflie position
 * Max threshold for sensors
 * Speed factor that set's how fast the Crazyflie moves

The demo is ended by either closing the graph window.

For the example to run the following hardware is needed:
 * Crazyflie 2.0
 * Crazyradio PA
 * Flow deck
 * Multiranger deck
"""
import logging
import math
import sys
from time import time

import numpy as np
from vispy import scene
from vispy.scene import visuals
from vispy.scene.cameras import TurntableCamera

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.utils import uri_helper

try:
    from sip import setapi
    setapi('QVariant', 2)
    setapi('QString', 2)
except ImportError:
    pass

from PyQt6 import QtCore, QtWidgets

logging.basicConfig(level=logging.INFO)

URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

if len(sys.argv) > 1:
    URI = sys.argv[1]

# Enable plotting of Crazyflie
PLOT_CF = False
# Enable plotting of down sensor
PLOT_SENSOR_DOWN = False
# Set the sensor threshold (in mm)
SENSOR_TH = 2000
# Set the speed factor for moving and rotating
SPEED_FACTOR = 0.3


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, URI):
        QtWidgets.QMainWindow.__init__(self)

        self.resize(700, 500)
        self.setWindowTitle('Multi-ranger point cloud')

        self.canvas = Canvas(self.updateHover)
        self.canvas.create_native()
        self.canvas.native.setParent(self)

        self.setCentralWidget(self.canvas.native)

        cflib.crtp.init_drivers()
        self.cf = Crazyflie(ro_cache=None, rw_cache='cache')

        # Connect callbacks from the Crazyflie API
        self.cf.connected.add_callback(self.connected)
        self.cf.disconnected.add_callback(self.disconnected)

        # Connect to the Crazyflie
        self.cf.open_link(URI)

        # Arm the Crazyflie
        self.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        self.hover = {'x': 0.0, 'y': 0.0, 'z': 0.0, 'yaw': 0.0, 'height': 0.3}

        self.hoverTimer = QtCore.QTimer()
        self.hoverTimer.timeout.connect(self.sendHoverCommand)
        self.hoverTimer.setInterval(100)
        self.hoverTimer.start()

    def sendHoverCommand(self):
        self.cf.commander.send_hover_setpoint(
            self.hover['x'], self.hover['y'], self.hover['yaw'],
            self.hover['height'])

    def updateHover(self, k, v):
        if (k != 'height'):
            self.hover[k] = v * SPEED_FACTOR
        else:
            self.hover[k] += v

    def disconnected(self, URI):
        print('Disconnected')

    def connected(self, URI):
        print('We are now connected to {}'.format(URI))

        # The definition of the logconfig can be made before connecting
        lpos = LogConfig(name='Position', period_in_ms=100)
        lpos.add_variable('stateEstimate.x')
        lpos.add_variable('stateEstimate.y')
        lpos.add_variable('stateEstimate.z')

        try:
            self.cf.log.add_config(lpos)
            lpos.data_received_cb.add_callback(self.pos_data)
            lpos.start()
        except KeyError as e:
            print('Could not start log configuration,'
                  '{} not found in TOC'.format(str(e)))
        except AttributeError:
            print('Could not add Position log config, bad configuration.')

        lmeas = LogConfig(name='Meas', period_in_ms=100)
        lmeas.add_variable('range.front')
        lmeas.add_variable('range.back')
        lmeas.add_variable('range.up')
        lmeas.add_variable('range.left')
        lmeas.add_variable('range.right')
        lmeas.add_variable('range.zrange')
        lmeas.add_variable('stabilizer.roll')
        lmeas.add_variable('stabilizer.pitch')
        lmeas.add_variable('stabilizer.yaw')

        try:
            self.cf.log.add_config(lmeas)
            lmeas.data_received_cb.add_callback(self.meas_data)
            lmeas.start()
        except KeyError as e:
            print('Could not start log configuration,'
                  '{} not found in TOC'.format(str(e)))
        except AttributeError:
            print('Could not add Measurement log config, bad configuration.')

    def pos_data(self, timestamp, data, logconf):
        position = [
            data['stateEstimate.x'],
            data['stateEstimate.y'],
            data['stateEstimate.z']
        ]
        self.canvas.set_position(position)

    def meas_data(self, timestamp, data, logconf):
        measurement = {
            'roll': data['stabilizer.roll'],
            'pitch': data['stabilizer.pitch'],
            'yaw': data['stabilizer.yaw'],
            'front': data['range.front'],
            'back': data['range.back'],
            'up': data['range.up'],
            'down': data['range.zrange'],
            'left': data['range.left'],
            'right': data['range.right']
        }
        self.canvas.set_measurement(measurement)

    def closeEvent(self, event):
        if (self.cf is not None):
            self.cf.close_link()


class Canvas(scene.SceneCanvas):
    def __init__(self, keyupdateCB):
        scene.SceneCanvas.__init__(self, keys=None)
        self.size = 800, 600
        self.unfreeze()
        self.view = self.central_widget.add_view()
        self.view.bgcolor = '#ffffff'
        self.view.camera = TurntableCamera(
            fov=10.0, distance=30.0, up='+z', center=(0.0, 0.0, 0.0))
        self.last_pos = [0, 0, 0]
        self.pos_markers = visuals.Markers()
        self.meas_markers = visuals.Markers()
        self.pos_data = np.array([0, 0, 0], ndmin=2)
        self.meas_data = np.array([0, 0, 0], ndmin=2)
        self.lines = []

        self.view.add(self.pos_markers)
        self.view.add(self.meas_markers)
        for i in range(6):
            line = visuals.Line()
            self.lines.append(line)
            self.view.add(line)

        self.keyCB = keyupdateCB

        self.freeze()

        scene.visuals.XYZAxis(parent=self.view.scene)

    def on_key_press(self, event):
        if (not event.native.isAutoRepeat()):
            if (event.native.key() == QtCore.Qt.Key.Key_Left):
                self.keyCB('y', 1)
            if (event.native.key() == QtCore.Qt.Key.Key_Right):
                self.keyCB('y', -1)
            if (event.native.key() == QtCore.Qt.Key.Key_Up):
                self.keyCB('x', 1)
            if (event.native.key() == QtCore.Qt.Key.Key_Down):
                self.keyCB('x', -1)
            if (event.native.key() == QtCore.Qt.Key.Key_A):
                self.keyCB('yaw', -70)
            if (event.native.key() == QtCore.Qt.Key.Key_D):
                self.keyCB('yaw', 70)
            if (event.native.key() == QtCore.Qt.Key.Key_Z):
                self.keyCB('yaw', -200)
            if (event.native.key() == QtCore.Qt.Key.Key_X):
                self.keyCB('yaw', 200)
            if (event.native.key() == QtCore.Qt.Key.Key_W):
                self.keyCB('height', 0.1)
            if (event.native.key() == QtCore.Qt.Key.Key_S):
                self.keyCB('height', -0.1)

    def on_key_release(self, event):
        if (not event.native.isAutoRepeat()):
            if (event.native.key() == QtCore.Qt.Key.Key_Left):
                self.keyCB('y', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_Right):
                self.keyCB('y', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_Up):
                self.keyCB('x', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_Down):
                self.keyCB('x', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_A):
                self.keyCB('yaw', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_D):
                self.keyCB('yaw', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_W):
                self.keyCB('height', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_S):
                self.keyCB('height', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_Z):
                self.keyCB('yaw', 0)
            if (event.native.key() == QtCore.Qt.Key.Key_X):
                self.keyCB('yaw', 0)

    def set_position(self, pos):
        self.last_pos = pos
        if (PLOT_CF):
            self.pos_data = np.append(self.pos_data, [pos], axis=0)
            self.pos_markers.set_data(self.pos_data, face_color='red', size=5)

    def rot(self, roll, pitch, yaw, origin, point):
        cosr = math.cos(math.radians(roll))
        cosp = math.cos(math.radians(pitch))
        cosy = math.cos(math.radians(yaw))

        sinr = math.sin(math.radians(roll))
        sinp = math.sin(math.radians(pitch))
        siny = math.sin(math.radians(yaw))

        roty = np.array([[cosy, -siny, 0],
                         [siny, cosy, 0],
                         [0, 0,    1]])

        rotp = np.array([[cosp, 0, sinp],
                         [0, 1, 0],
                         [-sinp, 0, cosp]])

        rotr = np.array([[1, 0,   0],
                         [0, cosr, -sinr],
                         [0, sinr,  cosr]])

        rotFirst = np.dot(rotr, rotp)

        rot = np.array(np.dot(rotFirst, roty))

        tmp = np.subtract(point, origin)
        tmp2 = np.dot(rot, tmp)
        return np.add(tmp2, origin)

    def rotate_and_create_points(self, m):
        data = []
        o = self.last_pos
        roll = m['roll']
        pitch = -m['pitch']
        yaw = m['yaw']

        if (m['up'] < SENSOR_TH):
            up = [o[0], o[1], o[2] + m['up'] / 1000.0]
            data.append(self.rot(roll, pitch, yaw, o, up))

        if (m['down'] < SENSOR_TH and PLOT_SENSOR_DOWN):
            down = [o[0], o[1], o[2] - m['down'] / 1000.0]
            data.append(self.rot(roll, pitch, yaw, o, down))

        if (m['left'] < SENSOR_TH):
            left = [o[0], o[1] + m['left'] / 1000.0, o[2]]
            data.append(self.rot(roll, pitch, yaw, o, left))

        if (m['right'] < SENSOR_TH):
            right = [o[0], o[1] - m['right'] / 1000.0, o[2]]
            data.append(self.rot(roll, pitch, yaw, o, right))

        if (m['front'] < SENSOR_TH):
            front = [o[0] + m['front'] / 1000.0, o[1], o[2]]
            data.append(self.rot(roll, pitch, yaw, o, front))

        if (m['back'] < SENSOR_TH):
            back = [o[0] - m['back'] / 1000.0, o[1], o[2]]
            data.append(self.rot(roll, pitch, yaw, o, back))

        return data

    def set_measurement(self, measurements):
        data = self.rotate_and_create_points(measurements)
        o = self.last_pos
        for i in range(6):
            if (i < len(data)):
                o = self.last_pos
                self.lines[i].set_data(np.array([o, data[i]]))
            else:
                self.lines[i].set_data(np.array([o, o]))

        if (len(data) > 0):
            self.meas_data = np.append(self.meas_data, data, axis=0)
        self.meas_markers.set_data(self.meas_data, face_color='blue', size=5)


if __name__ == '__main__':
    appQt = QtWidgets.QApplication(sys.argv)
    win = MainWindow(URI)
    win.show()
    appQt.exec()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2019 - 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import logging
import math
import struct

from .memory_element import MemoryElement
from cflib.utils.callbacks import Syncer

logger = logging.getLogger(__name__)


class Poly4D:
    class Poly:
        def __init__(self, values=[0.0] * 8):
            self.values = values

    def __init__(self, duration, x=None, y=None, z=None, yaw=None):
        self.duration = duration
        self.x = x if x else self.Poly()
        self.y = y if y else self.Poly()
        self.z = z if z else self.Poly()
        self.yaw = yaw if yaw else self.Poly()

    def pack(self):
        data = bytearray()

        data += struct.pack('<ffffffff', *self.x.values)
        data += struct.pack('<ffffffff', *self.y.values)
        data += struct.pack('<ffffffff', *self.z.values)
        data += struct.pack('<ffffffff', *self.yaw.values)
        data += struct.pack('<f', self.duration)

        return data


class _CompressedBase:
    def _encode_spatial(self, coordinate):
        '''
        Spatial coordinates (X, Y and Z) are represented as millimeters and are stored as signed 2-byte integers,
        meaning that the maximum spatial volume that this representation can cover is roughly 64m x 64m x 64m,
        assuming that the origin is at the center.
        '''
        return int(coordinate * 1000)

    def _encode_spatial_element(self, element):
        return map(self._encode_spatial, element)

    def _encode_yaw(self, angle_rad):
        '''
        Angles (for the yaw coordinate) are represented as 1/10th of degrees and are stored as signed 2-byte
        integers.
        '''
        return int(math.degrees(angle_rad) * 10)

    def _encode_yaw_element(self, element):
        '''
        Angles (for the yaw coordinate) are represented as 1/10th of degrees and are stored as signed 2-byte
        integers.
        '''
        return map(self._encode_yaw, element)


class CompressedStart(_CompressedBase):
    def __init__(self, x, y, z, yaw) -> None:
        self.x = x
        self.y = y
        self.z = z
        self.yaw = yaw

    def pack(self):
        data = bytearray()

        data += struct.pack(
            '<hhhh',
            self._encode_spatial(self.x),
            self._encode_spatial(self.y),
            self._encode_spatial(self.z),
            self._encode_yaw(self.yaw))

        return data


class CompressedSegment(_CompressedBase):
    def __init__(self, duration, element_x, element_y, element_z, element_yaw) -> None:
        self._validate(element_x)
        self._validate(element_y)
        self._validate(element_z)
        self._validate(element_yaw)

        self.duration = duration
        self.x = element_x
        self.y = element_y
        self.z = element_z
        self.yaw = element_yaw

    def pack(self):
        element_types = (self._encode_type(self.x) << 0) | (self._encode_type(self.y) << 2) | (
            self._encode_type(self.z) << 4) | (self._encode_type(self.yaw) << 6)
        duration_ms = int(self.duration * 1000.0)

        data = bytearray()

        data += struct.pack('<BH', element_types, duration_ms)
        data += self._pack_element(self._encode_spatial_element(self.x))
        data += self._pack_element(self._encode_spatial_element(self.y))
        data += self._pack_element(self._encode_spatial_element(self.z))
        data += self._pack_element(self._encode_yaw_element(self.yaw))

        return data

    def _validate(self, element):
        length = len(element)
        if length != 0 and length != 1 and length != 3 and length != 7:
            raise Exception('length of element must be 0, 1, 3, or 7')

    def _encode_type(self, element):
        if len(element) == 0:
            return 0
        if len(element) == 1:
            return 1
        if len(element) == 3:
            return 2
        if len(element) == 7:
            return 3

    def _pack_element(self, encoded_element):
        data = bytearray()

        for part in encoded_element:
            data += struct.pack('<h', part)

        return data


class TrajectoryMemory(MemoryElement):
    """
    Memory interface for trajectories used by the high level commander
    """

    def __init__(self, id, type, size, mem_handler):
        """Initialize trajectory memory"""
        super(TrajectoryMemory, self).__init__(id=id, type=type, size=size, mem_handler=mem_handler)
        self._write_finished_cb = None
        self._write_failed_cb = None

        # A list of trajectory elements to write to the Crazyflie. The elements can either be
        # Poly4D instances for uncompressed trajectories or one CompressedStart instance followed
        # by CompressedSegment instances. It is not possible to mix uncompressed and compressed
        # elements in the same trajectory.
        self.trajectory = []

    # Deprecated (removed after August 2023). replaced by self.trajectory
    @property
    def poly4Ds(self):
        return self.trajectory

    # Deprecated (removed after August 2023). replaced by self.trajectory
    @poly4Ds.setter
    def poly4Ds(self, trajectory):
        self.trajectory = trajectory

    def write_data(self, write_finished_cb, write_failed_cb=None, start_addr=0x00):
        """
        Write trajectory data to the Crazyflie.
        The trajectory in self.trajectory is written to the Crazyflie.
        By default the trajectory is written to address 0 of the Crazyflie trajectory memory, but it is possible to
        use a different address. This can be interesting if you want to define more than one trajectory but it requires
        careful handling of the addresses to avoid overwriting trajectories and staying within the trajectory memory.

        @param write_finished_cb A callback that is called when the write trajectory is uploaded.
        @param write_failed_cb Callback that is called if the upload failed
        @param start_addr The address in the trajectory memory to upload the trajectory to (0 by default)
        @return The number of bytes used for the trajectory
        """
        self._write_finished_cb = write_finished_cb
        self._write_failed_cb = write_failed_cb
        data = bytearray()

        for element in self.trajectory:
            data += element.pack()

        self.mem_handler.write(self, start_addr, data, flush_queue=True)
        return len(data)

    def write_data_sync(self, start_addr=0x00):
        """
        Same functionality as write_data() but synchronous (blocking)

        Args:
            start_addr (hexadecimal, optional): The address in the trajectory memory to upload the trajectory to.
            Defaults to 0x00.
        """
        syncer = Syncer()
        self.write_data(syncer.success_cb, write_failed_cb=syncer.failure_cb, start_addr=start_addr)
        syncer.wait()
        return syncer.is_success

    def write_done(self, mem, addr):
        if self._write_finished_cb and mem.id == self.id:
            logger.debug('Write trajectory data done')
            self._write_finished_cb(self, addr)
            self._write_finished_cb = None
            self._write_failed_cb = None

    def write_failed(self, mem, addr):
        if mem.id == self.id:
            if self._write_failed_cb:
                logger.debug('Write of trajectory data failed')
                self._write_failed_cb(self, addr)
            self._write_finished_cb = None
            self._write_failed_cb = None

    def disconnect(self):
        self._write_finished_cb = None



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2025 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')


class wrgb:
    def __init__(self, w: int, r: int, g: int, b: int):
        self.w = w
        self.r = r
        self.g = g
        self.b = b


class rgb:
    def __init__(self, r: int, g: int, b: int):
        self.w = min(r, g, b)
        self.r = r - self.w
        self.g = g - self.w
        self.b = b - self.w


def pack_wrgb(input: wrgb | rgb) -> int:
    """Pack WRGB values into uint32 format: 0xWWRRGGBB"""
    return (input.w << 24) | (input.r << 16) | (input.g << 8) | input.b


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf

        # Detect which color LED deck is present (bottom or top)
        # We check for whichever deck is attached and use the first one found
        # Check for bottom-facing deck first
        if str(cf.param.get_value('deck.bcColorLedBot')) == '1':
            deck_params = {
                'color': 'colorLedBot.wrgb8888',
                'brightness': 'colorLedBot.brightCorr',
                'throttle': 'colorLedBot.throttlePct',
                'temp': 'colorLedBot.deckTemp'
            }
            print('Detected bottom-facing color LED deck')
        # Check for top-facing deck
        elif str(cf.param.get_value('deck.bcColorLedTop')) == '1':
            deck_params = {
                'color': 'colorLedTop.wrgb8888',
                'brightness': 'colorLedTop.brightCorr',
                'throttle': 'colorLedTop.throttlePct',
                'temp': 'colorLedTop.deckTemp'
            }
            print('Detected top-facing color LED deck')
        else:
            raise RuntimeError('No color LED deck detected!')

        # Thermal status callback
        def thermal_status_callback(timestamp, data, logconf):
            throttle_pct = data[deck_params['throttle']]
            if throttle_pct > 0:
                temp = data[deck_params['temp']]
                print(f'WARNING: Thermal throttling active! Temp: {temp}°C, Throttle: {throttle_pct}%')

        # Setup log configuration for thermal monitoring
        log_conf = LogConfig(name='ThermalStatus', period_in_ms=100)
        log_conf.add_variable(deck_params['temp'], 'uint8_t')
        log_conf.add_variable(deck_params['throttle'], 'uint8_t')

        cf.log.add_config(log_conf)
        log_conf.data_received_cb.add_callback(thermal_status_callback)
        log_conf.start()

        # Brightness correction: balances luminance across R/G/B/W channels
        # Set to 1 (enabled, default) for perceptually uniform colors
        # Set to 0 (disabled) for maximum brightness per channel
        cf.param.set_value(deck_params['brightness'], 1)
        time.sleep(0.1)

        try:
            # ========================================
            # Set your desired color here:
            # ========================================
            # Examples:
            #   color = wrgb(w=0, r=255, g=0, b=0)      # Pure red
            #   color = wrgb(w=0, r=0, g=255, b=0)      # Pure green
            #   color = wrgb(w=0, r=0, g=0, b=255)      # Pure blue
            #   color = wrgb(w=255, r=0, g=0, b=0)      # Pure white LED
            #   color = wrgb(w=50, r=255, g=128, b=0)   # Orange + white
            #   color = rgb(r=255, g=255, b=255)        # White (auto W extraction)

            color = wrgb(w=0, r=255, g=0, b=100)
            # ========================================

            color_uint32 = pack_wrgb(color)
            cf.param.set_value(deck_params['color'], color_uint32)
            time.sleep(0.01)
            print(f'Setting LED to R={color.r}, G={color.g}, B={color.b}, W={color.w}')
            print('Press Ctrl-C to turn off LED and exit.')
            while True:
                time.sleep(0.1)
        except KeyboardInterrupt:
            print('\nStopping and turning off LED...')
            cf.param.set_value(deck_params['color'], 0)
            time.sleep(0.1)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import struct
import time
import unittest

import numpy as np
from single_cf_grounded import TestSingleCfGrounded

import cflib.crtp
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort


class TestLink(TestSingleCfGrounded):

    # def test_scan(self):
    #     start_time = time.time()
    #     result = cflib.crtp.scan_interfaces()
    #     end_time = time.time()
    #     uris = [uri for (uri, msg) in result]
    #     self.assertEqual(len(uris), 2)
    #     self.assertIn("radio://*/80/2M/E7E7E7E7E7", uris)
    #     self.assertIn("usb://0", uris)
    #     self.assertLess(end_time - start_time, 2)

    # def test_latency_radio_s4(self):
    #     result = self.latency(self.radioUri, 4)
    #     self.assertLess(result, 8)

    # def test_latency_radio_s28(self):
    #     result = self.latency(self.radioUri, 28)
    #     self.assertLess(result, 8)

    def test_latency_usb_s4(self):
        result = self.latency(self.usbUri, 4, 1000)
        self.assertLess(result, 1)

    def test_latency_usb_s28(self):
        result = self.latency(self.usbUri, 28, 1000)
        self.assertLess(result, 1)

    # def test_bandwidth_radio_s4(self):
    #     result = self.bandwidth(self.radioUri, 4)
    #     self.assertGreater(result, 450)

    # def test_bandwidth_radio_s28(self):
    #     result = self.bandwidth(self.radioUri, 28)
    #     self.assertGreater(result, 450)

    def test_bandwidth_usb_s4(self):
        result = self.bandwidth(self.usbUri, 4, 1000)
        self.assertGreater(result, 1000)

    def test_bandwidth_usb_s28(self):
        result = self.bandwidth(self.usbUri, 28, 1000)
        self.assertGreater(result, 1500)

    def latency(self, uri, packet_size=4, count=500):
        link = cflib.crtp.get_link_driver(uri)
        # # wait until no more packets in queue
        # while True:
        #     pk = link.receive_packet(0.5)
        #     print(pk)
        #     if not pk or pk.header == 0xFF:
        #         break

        pk = CRTPPacket()
        pk.set_header(CRTPPort.LINKCTRL, 0)  # Echo channel

        latencies = []
        for i in range(count):
            pk.data = self.build_data(i, packet_size)

            start_time = time.time()
            link.send_packet(pk)
            while True:
                pk_ack = link.receive_packet(-1)
                if pk_ack.port == CRTPPort.LINKCTRL and pk_ack.channel == 0:
                    break
            end_time = time.time()

            # make sure we actually received the expected value
            i_recv, = struct.unpack('<I', pk_ack.data[0:4])
            self.assertEqual(i, i_recv)
            latencies.append((end_time - start_time) * 1000)
        link.close()
        result = np.min(latencies)
        print('Latency for {} (packet size {} B): {:.2f} ms'.format(uri, packet_size, result))
        return result

    def bandwidth(self, uri, packet_size=4, count=500):
        link = cflib.crtp.get_link_driver(uri, link_error_callback=self.error_cb)
        # # wait until no more packets in queue
        # while True:
        #     pk = link.receive_packet(0.5)
        #     if not pk:
        #         break

        # enqueue packets
        start_time = time.time()
        for i in range(count):
            pk = CRTPPacket()
            pk.set_header(CRTPPort.LINKCTRL, 0)  # Echo channel
            pk.data = self.build_data(i, packet_size)
            link.send_packet(pk)

        # get the result
        for i in range(count):
            while True:
                pk_ack = link.receive_packet(-1)
                if pk_ack.port == CRTPPort.LINKCTRL and pk_ack.channel == 0:
                    break
            # make sure we actually received the expected value
            i_recv, = struct.unpack('<I', pk_ack.data[0:4])
            self.assertEqual(i, i_recv)
        end_time = time.time()
        link.close()
        result = count / (end_time - start_time)
        kbps = (count * packet_size) / 1024 / (end_time - start_time)
        print('Bandwith for {} (packet size {} B): {:.2f} packets/s ({:.2f} kB/s)'.format(
            uri, packet_size, result, kbps))
        return result

    def error_cb(self, error):
        self.assertIsNone(None, error)

    def build_data(self, i, packet_size):
        assert (packet_size % 4 == 0)
        repeats = packet_size // 4
        return struct.pack('<' + 'I'*repeats, *[i]*repeats)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021-2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import math

import numpy as np
import numpy.typing as npt


class LighthouseBsVector:
    """
    This class is representing a vector from a base station into space,
    in the base station reference frame. Typically the intersection of
    two light planes defined by angles measured by a base station.
    It also provides functionality to convert between lighthouse V1 angles,
    V2 angles and cartesian coordinates.
    """

    T = math.pi / 6

    def __init__(self, lh_v1_horiz_angle: float, lh_v1_vert_angle: float) -> None:
        """
        Initialize from lighthouse V1 angles
        :param lh_v1_horiz_angle: Horizontal sweep angle, 0 straight forward. Right (seen from the bs) is negative,
                                  left is positive
        :param lh_v1_vert_angle: Vertical sweep angle, 0 straight forward. Down is negative, up is positive.
        """
        self._lh_v1_horiz_angle = lh_v1_horiz_angle
        self._lh_v1_vert_angle = lh_v1_vert_angle

    @classmethod
    def from_lh2(cls, lh_v2_angle_1: float, lh_v2_angle_2: float) -> 'LighthouseBsVector':
        """
        Create a LighthouseBsVector object from lighthouse V2 angles
        :param lh_v2_angle_1: First sweep angles, 0 straight ahead
        :param lh_v2_angle_2: Second sweep angles, 0 straight ahead
        """
        a1 = lh_v2_angle_1
        a2 = lh_v2_angle_2
        lh_v1_horiz_angle = (a1 + a2) / 2.0
        lh_v1_vert_angle = math.atan2(math.sin(a2 - a1), math.tan(cls.T) * (math.cos(a1) + math.cos(a2)))

        return cls(lh_v1_horiz_angle, lh_v1_vert_angle)

    @classmethod
    def from_cart(cls, cart_vector: list[float]) -> 'LighthouseBsVector':
        """
        Create a LighthouseBsVector object from cartesian coordinates.
        :param cart_vector: (x, y, z) to a point
        """
        lh_v1_horiz_angle = math.atan2(cart_vector[1], cart_vector[0])
        lh_v1_vert_angle = math.atan2(cart_vector[2], cart_vector[0])

        return cls(lh_v1_horiz_angle, lh_v1_vert_angle)

    @classmethod
    def from_projection(cls, proj_point: list[float]) -> 'LighthouseBsVector':
        """
        Create a LighthouseBsVector object from the projection point on the plane x=1.0
        :param projection point: (y, z)
        """
        lh_v1_horiz_angle = math.atan(proj_point[0])
        lh_v1_vert_angle = math.atan(proj_point[1])

        return cls(lh_v1_horiz_angle, lh_v1_vert_angle)

    @property
    def lh_v1_horiz_angle(self) -> float:
        """
        Lightouse V1 horizontal sweep angle
        """
        return self._lh_v1_horiz_angle

    @property
    def lh_v1_vert_angle(self) -> float:
        """
        Lightouse V1 vertical sweep angle
        """
        return self._lh_v1_vert_angle

    @property
    def lh_v1_angle_pair(self) -> tuple[float, float]:
        """
        Lightouse V1 angle pair (horiz, vert)
        """
        return self._lh_v1_horiz_angle, self._lh_v1_vert_angle,

    @property
    def lh_v2_angle_1(self) -> float:
        """
        Lightouse V2 first sweep angle
        """
        return self._lh_v1_horiz_angle + math.asin(self._q() * math.tan(-self.T))

    @property
    def lh_v2_angle_2(self) -> float:
        """
        Lightouse V2 second sweep angle
        """
        return self._lh_v1_horiz_angle + math.asin(self._q() * math.tan(self.T))

    @property
    def cart(self) -> npt.NDArray[np.float32]:
        """
        A normalized vector in cartesian coordinates
        """
        v = np.float32((1, math.tan(self._lh_v1_horiz_angle), math.tan(self._lh_v1_vert_angle)))
        return v / np.linalg.norm(v)

    @property
    def projection(self) -> npt.NDArray[np.float32]:
        """
        The 2D point (y, z) when projected on the plane x=1.0 (one meter in front of the base station)
        """
        return np.float32((math.tan(self._lh_v1_horiz_angle), math.tan(self._lh_v1_vert_angle)))

    def _q(self):
        return math.tan(self._lh_v1_vert_angle) / math.sqrt(1 + math.tan(self._lh_v1_horiz_angle) ** 2)


class LighthouseBsVectors(list):
    """A list of 4 LighthouseBsVector, one for each sensor.
       LighthouseBsVectors is essentially the same as list[LighthouseBsVector]"""

    def projection_pair_list(self) -> npt.NDArray:
        """
        Genereate a list of projection pairs for all vectors
        """
        result = np.empty((len(self), 2), dtype=float)
        for i, vector in enumerate(self):
            result[i] = vector.projection

        return result

    def angle_list(self) -> npt.NDArray:
        """
        Genereate a list of angles for all vectors, the order is horizontal, vertical, horizontal, vertical...
        """
        result = np.empty((len(self) * 2), dtype=float)
        for i, vector in enumerate(self):
            result[i * 2] = vector.lh_v1_horiz_angle
            result[i * 2 + 1] = vector.lh_v1_vert_angle

        return result



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example of how to read the memory from the multiranger
"""
import logging
import time
from threading import Event

import matplotlib.pyplot as plt

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class ReadMem:
    def __init__(self, uri):
        self._event = Event()
        self._cf = Crazyflie(rw_cache='./cache')

        with SyncCrazyflie(uri, cf=self._cf) as scf:
            mems = scf.cf.mem.get_mems(MemoryElement.TYPE_DECK_PAA3905)

            count = len(mems)
            if count != 1:
                raise Exception('Unexpected nr of memories found:', count)

            mem = mems[0]

            data = [[0 for x in range(35)] for y in range(35)]
            im = plt.imshow(data, cmap='gray', vmin=0, vmax=255, origin='upper')

            start_time = time.time()
            for frames in range(100):
                data = mem.read_data_sync()
                im.set_data(data)
                plt.pause(0.01)

            end_time = time.time()
            print('FPS: {}'.format(100/(end_time - start_time)))
            time.sleep(5)


if __name__ == '__main__':
    # URI to the Crazyflie to connect to
    #    uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')
    uri = uri_helper.uri_from_env(default='usb://0')

    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    rm = ReadMem(uri)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  Crazyflie Python Library
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example scripts that allows a user to "push" the Crazyflie 2.0 around
using your hands while it's hovering.

This examples uses the Flow and Multi-ranger decks to measure distances
in all directions and tries to keep away from anything that comes closer
than 0.2m by setting a velocity in the opposite direction.

The demo is ended by either pressing Ctrl-C or by holding your hand above the
Crazyflie.

For the example to run the following hardware is needed:
 * Crazyflie 2.0
 * Crazyradio PA
 * Flow deck
 * Multiranger deck
"""
import logging
import sys
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.positioning.motion_commander import MotionCommander
from cflib.utils import uri_helper
from cflib.utils.multiranger import Multiranger

URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

if len(sys.argv) > 1:
    URI = sys.argv[1]

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


def is_close(range):
    MIN_DISTANCE = 0.2  # m

    if range is None:
        return False
    else:
        return range < MIN_DISTANCE


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    cf = Crazyflie(rw_cache='./cache')
    with SyncCrazyflie(URI, cf=cf) as scf:
        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        with MotionCommander(scf) as motion_commander:
            with Multiranger(scf) as multiranger:
                keep_flying = True

                while keep_flying:
                    VELOCITY = 0.5
                    velocity_x = 0.0
                    velocity_y = 0.0

                    if is_close(multiranger.front):
                        velocity_x -= VELOCITY
                    if is_close(multiranger.back):
                        velocity_x += VELOCITY

                    if is_close(multiranger.left):
                        velocity_y -= VELOCITY
                    if is_close(multiranger.right):
                        velocity_y += VELOCITY

                    if is_close(multiranger.up):
                        keep_flying = False

                    motion_commander.start_linear_motion(
                        velocity_x, velocity_y, 0)

                    time.sleep(0.1)

            print('Demo terminated!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Enables access to the Crazyflie memory subsystem.

"""
import errno
import logging
import struct
from threading import Lock

from .deck_memory import DeckMemoryManager
from .i2c_element import I2CElement
from .led_driver_memory import LEDDriverMemory
from .led_timings_driver_memory import LEDTimingsDriverMemory
from .lighthouse_memory import LighthouseBsCalibration
from .lighthouse_memory import LighthouseBsGeometry
from .lighthouse_memory import LighthouseMemHelper
from .lighthouse_memory import LighthouseMemory
from .loco_memory import LocoMemory
from .loco_memory_2 import LocoMemory2
from .memory_element import MemoryElement
from .memory_tester import MemoryTester
from .multiranger_memory import MultirangerMemory
from .ow_element import OWElement
from .paa3905_memory import PAA3905Memory
from .trajectory_memory import CompressedSegment
from .trajectory_memory import CompressedStart
from .trajectory_memory import Poly4D
from .trajectory_memory import TrajectoryMemory
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Memory', 'Poly4D', 'CompressedStart', 'CompressedSegment', 'MemoryElement',
           'LighthouseBsGeometry', 'LighthouseBsCalibration', 'LighthouseMemHelper',
           'DeckMemoryManager']

# Channels used for the logging port
CHAN_INFO = 0
CHAN_READ = 1
CHAN_WRITE = 2

# Commands used when accessing the Settings port
CMD_INFO_VER = 0
CMD_INFO_NBR = 1
CMD_INFO_DETAILS = 2

logger = logging.getLogger(__name__)


class _ReadRequest:
    """
    Class used to handle memory reads that will split up the read in multiple
    packets if necessary
    """
    MAX_DATA_LENGTH = 24

    def __init__(self, mem, addr, length, cf):
        """Initialize the object with good defaults"""
        self.mem = mem
        self.addr = addr
        self._bytes_left = length
        self.data = bytearray()
        self.cf = cf

        self._current_addr = addr

    def start(self):
        """Start the fetching of the data"""
        self._request_new_chunk()

    def resend(self):
        logger.debug('Sending write again...')
        self._request_new_chunk()

    def _request_new_chunk(self):
        """
        Called to request a new chunk of data to be read from the Crazyflie
        """
        # Figure out the length of the next request
        new_len = self._bytes_left
        if new_len > _ReadRequest.MAX_DATA_LENGTH:
            new_len = _ReadRequest.MAX_DATA_LENGTH

        logger.debug('Requesting new chunk of {}bytes at 0x{:X}'.format(
            new_len, self._current_addr))

        # Request the data for the next address
        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_READ)
        pk.data = struct.pack('<BIB', self.mem.id, self._current_addr, new_len)
        reply = struct.unpack('<BBBBB', pk.data[:-1])
        self.cf.send_packet(pk, expected_reply=reply, timeout=1)

    def add_data(self, addr, data):
        """Callback when data is received from the Crazyflie"""
        data_len = len(data)
        if not addr == self._current_addr:
            logger.warning(
                'Address did not match when adding data to read request!')
            return

        # Add the data and calculate the next address to fetch
        self.data += data
        self._bytes_left -= data_len
        self._current_addr += data_len

        if self._bytes_left > 0:
            self._request_new_chunk()
            return False
        else:
            return True


class _WriteRequest:
    """
    Class used to handle memory reads that will split up the read in multiple
    packets in necessary
    """
    MAX_DATA_LENGTH = 24

    def __init__(self, mem, addr, data, cf, progress_cb=None):
        """Initialize the object with good defaults"""
        self.mem = mem
        self.addr = addr
        self._bytes_left = len(data)
        self._write_len = self._bytes_left
        self._data = data
        self.data = bytearray()
        self.cf = cf
        self._progress_cb = progress_cb
        self._progress = -1

        self._current_addr = addr

        self._sent_packet = None
        self._sent_reply = None

        self._addr_add = 0

    def start(self):
        """Start the fetching of the data"""
        self._write_new_chunk()

    def resend(self):
        logger.debug('Sending write again...')
        self.cf.send_packet(
            self._sent_packet, expected_reply=self._sent_reply, timeout=1)

    def _write_new_chunk(self):
        """
        Called to request a new chunk of data to be read from the Crazyflie
        """
        # Figure out the length of the next request
        new_len = len(self._data)
        if new_len > _WriteRequest.MAX_DATA_LENGTH:
            new_len = _WriteRequest.MAX_DATA_LENGTH

        logger.debug('Writing new chunk of {}bytes at 0x{:X}'.format(
            new_len, self._current_addr))

        data = self._data[:new_len]
        self._data = self._data[new_len:]

        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_WRITE)
        pk.data = struct.pack('<BI', self.mem.id, self._current_addr)
        # Create a tuple used for matching the reply using id and address
        reply = struct.unpack('<BBBBB', pk.data)
        self._sent_reply = reply
        # Add the data
        pk.data += struct.pack('B' * len(data), *data)
        self._sent_packet = pk
        self.cf.send_packet(pk, expected_reply=reply, timeout=1)

        self._addr_add = len(data)
        self._bytes_left -= self._addr_add

    def _get_progress_message(self):
        if isinstance(self.mem, DeckMemoryManager):
            for deck_memory in self.mem.deck_memories.values():
                if deck_memory.contains(self._current_addr):
                    return f'Writing to {deck_memory.name} deck memory'

        return 'Writing to memory'

    def write_done(self, addr):
        """Callback when data is received from the Crazyflie"""
        if not addr == self._current_addr:
            logger.warning(
                'Address did not match when adding data to read request!')
            return

        if self._progress_cb is not None:
            new_progress = int(100 * (self._write_len - self._bytes_left) / self._write_len)
            if new_progress > self._progress:
                self._progress = new_progress
                self._progress_cb(self._get_progress_message(), self._progress)

        if len(self._data) > 0:
            self._current_addr += self._addr_add
            self._write_new_chunk()
            return False
        else:
            logger.debug('This write request is done')
            return True


class Memory():
    """Access memories on the Crazyflie"""

    # These codes can be decoded using os.stderror, but
    # some of the text messages will look very strange
    # in the UI, so they are redefined here
    _err_codes = {
        errno.ENOMEM: 'No more memory available',
        errno.ENOEXEC: 'Command not found',
        errno.ENOENT: 'No such block id',
        errno.E2BIG: 'Block too large',
        errno.EEXIST: 'Block already exists'
    }

    def __init__(self, crazyflie=None):
        """Instantiate class and connect callbacks"""
        self.cf = crazyflie
        self.cf.add_port_callback(CRTPPort.MEM, self._new_packet_cb)
        self.cf.disconnected.add_callback(self._disconnected)
        self._write_requests_lock = Lock()

        self._clear_state()

    def _clear_state(self):
        self.mems = []
        # Called when new memories have been added
        self.mem_added_cb = Caller()

        self._clear_refresh_callbacks()

        # Called to signal completion of read or write
        self.mem_read_cb = Caller()
        self.mem_read_failed_cb = Caller()
        self.mem_write_cb = Caller()
        self.mem_write_failed_cb = Caller()

        self._refresh_callback = None
        self._refresh_failed_callback = None
        self._fetch_id = 0
        self.nbr_of_mems = 0
        self._ow_mem_fetch_index = 0
        self._elem_data = ()
        self._read_requests = {}
        self._write_requests = {}
        self._ow_mems_left_to_update = []
        self._getting_count = False

    def _clear_refresh_callbacks(self):
        self._refresh_callback = None
        self._refresh_failed_callback = None

    def _mem_update_done(self, mem):
        """
        Callback from each individual memory (only 1-wire) when reading of
        header/elements are done
        """
        if mem.id in self._ow_mems_left_to_update:
            self._ow_mems_left_to_update.remove(mem.id)

        logger.debug(mem)

        if len(self._ow_mems_left_to_update) == 0:
            if self._refresh_callback:
                self._refresh_callback()
                self._clear_refresh_callbacks()

    def get_mem(self, id):
        """Fetch the memory with the supplied id"""
        for m in self.mems:
            if m.id == id:
                return m

        return None

    def get_mems(self, type):
        """Fetch all the memories of the supplied type"""
        ret = ()
        for m in self.mems:
            if m.type == type:
                ret += (m,)

        return ret

    def ow_search(self, vid=0xBC, pid=None, name=None):
        """Search for specific memory id/name and return it"""
        for m in self.get_mems(MemoryElement.TYPE_1W):
            if pid and m.pid == pid or name and m.name == name:
                return m

        return None

    def write(self, memory, addr, data, flush_queue=False, progress_cb=None):
        """Write the specified data to the given memory at the given address"""
        wreq = _WriteRequest(memory, addr, data, self.cf, progress_cb)
        if memory.id not in self._write_requests:
            self._write_requests[memory.id] = []

        # Workaround until we secure the uplink and change messages for
        # mems to non-blocking
        self._write_requests_lock.acquire()
        if flush_queue:
            self._write_requests[memory.id] = self._write_requests[
                memory.id][:1]
        self._write_requests[memory.id].append(wreq)
        if len(self._write_requests[memory.id]) == 1:
            wreq.start()
        self._write_requests_lock.release()

        return True

    def read(self, memory, addr, length):
        """
        Read the specified amount of bytes from the given memory at the given address
        """
        if memory.id in self._read_requests:
            logger.warning('There is already a read operation ongoing for memory id {}'.format(memory.id))
            return False

        rreq = _ReadRequest(memory, addr, length, self.cf)
        self._read_requests[memory.id] = rreq

        rreq.start()

        return True

    def refresh(self, refresh_done_callback, refresh_failed_cb=None):
        """Start fetching all the detected memories"""
        self._refresh_callback = refresh_done_callback
        self._refresh_failed_callback = refresh_failed_cb
        self._fetch_id = 0
        for m in self.mems:
            try:
                self.mem_read_cb.remove_callback(m.new_data)
                m.disconnect()
            except Exception as e:
                logger.info('Error when removing memory after update: {}'.format(e))
        self.mems = []

        self.nbr_of_mems = 0
        self._getting_count = False

        logger.debug('Requesting number of memories')
        pk = CRTPPacket()
        pk.set_header(CRTPPort.MEM, CHAN_INFO)
        pk.data = (CMD_INFO_NBR,)
        self.cf.send_packet(pk, expected_reply=(CMD_INFO_NBR,))

    def _disconnected(self, uri):
        """The link to the Crazyflie has been broken. Reset state"""
        self._call_all_failed_callbacks()
        self._clear_state()

    def _call_all_failed_callbacks(self):
        # Read requests
        read_requests = list(self._read_requests.values())
        self._read_requests.clear()
        for rreq in read_requests:
            self.mem_read_failed_cb.call(rreq.mem, rreq.addr, rreq.data)

        # Write requests
        write_requests = []
        self._write_requests_lock.acquire()
        for requests in self._write_requests.values():
            write_requests += requests
        self._write_requests.clear()
        self._write_requests_lock.release()

        for wreq in write_requests:
            self.mem_write_failed_cb.call(wreq.mem, wreq.addr)

        # Info
        if self._refresh_failed_callback:
            self._refresh_failed_callback()
            self._clear_refresh_callbacks()

    def _new_packet_cb(self, packet):
        """Callback for newly arrived packets for the memory port"""
        chan = packet.channel
        cmd = packet.data[0]
        payload = packet.data[1:]

        if chan == CHAN_INFO:
            self._handle_chan_info(cmd, payload)
        if chan == CHAN_WRITE:
            self._handle_chan_write(cmd, payload)
        if chan == CHAN_READ:
            self._handle_chan_read(cmd, payload)

    def _handle_chan_info(self, cmd, payload):
        if cmd == CMD_INFO_NBR:
            self._handle_cmd_info_nbr(payload)
        if cmd == CMD_INFO_DETAILS:
            self._handle_cmd_info_details(payload)

    def _handle_cmd_info_nbr(self, payload):
        self.nbr_of_mems = payload[0]
        logger.info('{} memories found'.format(self.nbr_of_mems))

        # Start requesting information about the memories,
        if self.nbr_of_mems > 0:
            if not self._getting_count:
                self._getting_count = True
                logger.debug('Requesting first id')
                pk = CRTPPacket()
                pk.set_header(CRTPPort.MEM, CHAN_INFO)
                pk.data = (CMD_INFO_DETAILS, 0)
                self.cf.send_packet(pk, expected_reply=(CMD_INFO_DETAILS, 0))
        else:
            if self._refresh_callback:
                self._refresh_callback()
                self._clear_refresh_callbacks()

    def _handle_cmd_info_details(self, payload):
        # Did we get a good reply, otherwise try again:
        if len(payload) < 5:
            # Workaround for 1-wire bug when memory is detected
            # but updating the info crashes the communication with
            # the 1-wire. Fail by saying we only found 1 memory
            # (the I2C).
            logger.error('-------->Got good count, but no info on mem!')
            self.nbr_of_mems = 1
            if self._refresh_callback:
                self._refresh_callback()
                self._clear_refresh_callbacks()
            return

        # Create information about a new memory
        # Id - 1 byte
        mem_id = payload[0]
        # Type - 1 byte
        mem_type = payload[1]
        # Size 4 bytes (as addr)
        mem_size = struct.unpack('I', payload[2:6])[0]
        # Addr (only valid for 1-wire?)
        mem_addr_raw = struct.unpack('B' * 8, payload[6:14])
        mem_addr = ''
        for m in mem_addr_raw:
            mem_addr += '{:02X}'.format(m)

        if (not self.get_mem(mem_id)):
            if mem_type == MemoryElement.TYPE_1W:
                mem = OWElement(id=mem_id, type=mem_type,
                                size=mem_size,
                                addr=mem_addr, mem_handler=self)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_write_cb.add_callback(mem.write_done)
                self._ow_mems_left_to_update.append(mem.id)
            elif mem_type == MemoryElement.TYPE_I2C:
                mem = I2CElement(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_write_cb.add_callback(mem.write_done)
            elif mem_type == MemoryElement.TYPE_DRIVER_LED:
                mem = LEDDriverMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_write_cb.add_callback(mem.write_done)
            elif mem_type == MemoryElement.TYPE_LOCO:
                mem = LocoMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
            elif mem_type == MemoryElement.TYPE_TRAJ:
                mem = TrajectoryMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_write_cb.add_callback(mem.write_done)
                self.mem_write_failed_cb.add_callback(mem.write_failed)
            elif mem_type == MemoryElement.TYPE_LOCO2:
                mem = LocoMemory2(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
            elif mem_type == MemoryElement.TYPE_LH:
                mem = LighthouseMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_read_failed_cb.add_callback(mem.new_data_failed)
                self.mem_write_cb.add_callback(mem.write_done)
                self.mem_write_failed_cb.add_callback(mem.write_failed)
            elif mem_type == MemoryElement.TYPE_MEMORY_TESTER:
                mem = MemoryTester(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_write_cb.add_callback(mem.write_done)
            elif mem_type == MemoryElement.TYPE_DRIVER_LEDTIMING:
                mem = LEDTimingsDriverMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_write_cb.add_callback(mem.write_done)
            elif mem_type == MemoryElement.TYPE_DECK_MEMORY:
                mem = DeckMemoryManager(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem._new_data)
                self.mem_read_failed_cb.add_callback(mem._new_data_failed)
                self.mem_write_cb.add_callback(mem._write_done)
                self.mem_write_failed_cb.add_callback(mem._write_failed)
            elif mem_type == MemoryElement.TYPE_DECK_MULTIRANGER:
                mem = MultirangerMemory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_read_failed_cb.add_callback(mem.read_failed)
            elif mem_type == MemoryElement.TYPE_DECK_PAA3905:
                mem = PAA3905Memory(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
                self.mem_read_cb.add_callback(mem.new_data)
                self.mem_read_failed_cb.add_callback(mem.read_failed)
            else:
                mem = MemoryElement(id=mem_id, type=mem_type, size=mem_size, mem_handler=self)
                logger.debug(mem)
            self.mems.append(mem)
            self.mem_added_cb.call(mem)

            self._fetch_id = mem_id + 1

        if self.nbr_of_mems - 1 >= self._fetch_id:
            logger.debug('Requesting information about memory {}'.format(self._fetch_id))
            pk = CRTPPacket()
            pk.set_header(CRTPPort.MEM, CHAN_INFO)
            pk.data = (CMD_INFO_DETAILS, self._fetch_id)
            self.cf.send_packet(pk, expected_reply=(CMD_INFO_DETAILS, self._fetch_id))
        else:
            logger.debug('Done getting all the memories, start reading the OWs')
            ows = self.get_mems(MemoryElement.TYPE_1W)
            # If there are any OW mems start reading them, otherwise
            # we are done
            for ow_mem in ows:
                ow_mem.update(self._mem_update_done)
            if len(ows) == 0:
                if self._refresh_callback:
                    self._refresh_callback()
                    self._clear_refresh_callbacks()

    def _handle_chan_write(self, cmd, payload):
        id = cmd
        (addr, status) = struct.unpack('<IB', payload[0:5])
        logger.debug('WRITE: Mem={}, addr=0x{:X}, status=0x{}'.format(id, addr, status))
        # Find the write request
        if id in self._write_requests:
            self._write_requests_lock.acquire()
            do_call_success_cb = False
            do_call_fail_cb = False
            wreq = self._write_requests[id][0]
            if status == 0:
                if wreq.write_done(addr):
                    # self._write_requests.pop(id, None)
                    # Remove the first item
                    self._write_requests[id].pop(0)
                    do_call_success_cb = True

                    # Get a new one to start (if there are any)
                    if len(self._write_requests[id]) > 0:
                        self._write_requests[id][0].start()
            else:
                logger.debug('Status {}: write failed.'.format(status))
                # Remove from queue
                self._write_requests[id].pop(0)
                do_call_fail_cb = True

                # Get a new one to start (if there are any)
                if len(self._write_requests[id]) > 0:
                    self._write_requests[id][0].start()

            self._write_requests_lock.release()

            # Call callbacks after the lock has been released to allow for new writes
            # to be initiated from the callback.
            if do_call_success_cb:
                self.mem_write_cb.call(wreq.mem, wreq.addr)
            if do_call_fail_cb:
                self.mem_write_failed_cb.call(wreq.mem, wreq.addr)

    def _handle_chan_read(self, cmd, payload):
        id = cmd
        (addr, status) = struct.unpack('<IB', payload[0:5])
        data = struct.unpack('B' * len(payload[5:]), payload[5:])
        logger.debug('READ: Mem={}, addr=0x{:X}, status=0x{}, data={}'.format(id, addr, status, data))
        # Find the read request
        if id in self._read_requests:
            logger.debug('READING: We are still interested in request for mem {}'.format(id))
            rreq = self._read_requests[id]
            if status == 0:
                if rreq.add_data(addr, payload[5:]):
                    self._read_requests.pop(id, None)
                    self.mem_read_cb.call(rreq.mem, rreq.addr, rreq.data)
            else:
                logger.debug('Status {}: read failed.'.format(status))
                self._read_requests.pop(id, None)
                self.mem_read_failed_cb.call(rreq.mem, rreq.addr, rreq.data)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017-2018 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Version of the AutonomousSequence.py example connecting to 10 Crazyflies.
The Crazyflies go straight up, hover a while and land but the code is fairly
generic and each Crazyflie has its own sequence of setpoints that it files
to.

The layout of the positions:
    x2      x1      x0

y3  10              4

            ^ Y
            |
y2  9       6       3
            |
            +------> X

y1  8       5       2



y0  7               1

"""
import time

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm

# Change uris and sequences according to your setup
# URIs in a swarm using the same radio must also be on the same channel
URI1 = 'radio://0/70/2M/E7E7E7E701'
URI2 = 'radio://0/70/2M/E7E7E7E702'
URI3 = 'radio://0/70/2M/E7E7E7E703'
URI4 = 'radio://0/70/2M/E7E7E7E704'
URI5 = 'radio://0/70/2M/E7E7E7E705'
URI6 = 'radio://0/70/2M/E7E7E7E706'
URI7 = 'radio://0/70/2M/E7E7E7E707'
URI8 = 'radio://0/70/2M/E7E7E7E708'
URI9 = 'radio://0/70/2M/E7E7E7E709'
URI10 = 'radio://0/70/2M/E7E7E7E70A'


z0 = 0.4
z = 1.0

x0 = 0.7
x1 = 0
x2 = -0.7

y0 = -1.0
y1 = -0.4
y2 = 0.4
y3 = 1.0

#    x   y   z  time
sequence1 = [
    (x0, y0, z0, 3.0),
    (x0, y0, z, 30.0),
    (x0, y0, z0, 3.0),
]

sequence2 = [
    (x0, y1, z0, 3.0),
    (x0, y1, z, 30.0),
    (x0, y1, z0, 3.0),
]

sequence3 = [
    (x0, y2, z0, 3.0),
    (x0, y2, z, 30.0),
    (x0, y2, z0, 3.0),
]

sequence4 = [
    (x0, y3, z0, 3.0),
    (x0, y3, z, 30.0),
    (x0, y3, z0, 3.0),
]

sequence5 = [
    (x1, y1, z0, 3.0),
    (x1, y1, z, 30.0),
    (x1, y1, z0, 3.0),
]

sequence6 = [
    (x1, y2, z0, 3.0),
    (x1, y2, z, 30.0),
    (x1, y2, z0, 3.0),
]

sequence7 = [
    (x2, y0, z0, 3.0),
    (x2, y0, z, 30.0),
    (x2, y0, z0, 3.0),
]

sequence8 = [
    (x2, y1, z0, 3.0),
    (x2, y1, z, 30.0),
    (x2, y1, z0, 3.0),
]

sequence9 = [
    (x2, y2, z0, 3.0),
    (x2, y2, z, 30.0),
    (x2, y2, z0, 3.0),
]

sequence10 = [
    (x2, y3, z0, 3.0),
    (x2, y3, z, 30.0),
    (x2, y3, z0, 3.0),
]

seq_args = {
    URI1: [sequence1],
    URI2: [sequence2],
    URI3: [sequence3],
    URI4: [sequence4],
    URI5: [sequence5],
    URI6: [sequence6],
    URI7: [sequence7],
    URI8: [sequence8],
    URI9: [sequence9],
    URI10: [sequence10],
}

# List of URIs, comment the one you do not want to fly
uris = {
    URI1,
    URI2,
    URI3,
    URI4,
    URI5,
    URI6,
    URI7,
    URI8,
    URI9,
    URI10
}


def wait_for_param_download(scf):
    while not scf.cf.param.is_updated:
        time.sleep(1.0)
    print('Parameters downloaded for', scf.cf.link_uri)


def arm(scf):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


def take_off(cf, position):
    take_off_time = 1.0
    sleep_time = 0.1
    steps = int(take_off_time / sleep_time)
    vz = position[2] / take_off_time

    print(vz)

    for i in range(steps):
        cf.commander.send_velocity_world_setpoint(0, 0, vz, 0)
        time.sleep(sleep_time)


def land(cf, position):
    landing_time = 1.0
    sleep_time = 0.1
    steps = int(landing_time / sleep_time)
    vz = -position[2] / landing_time

    print(vz)

    for _ in range(steps):
        cf.commander.send_velocity_world_setpoint(0, 0, vz, 0)
        time.sleep(sleep_time)

    cf.commander.send_stop_setpoint()
    # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
    cf.commander.send_notify_setpoint_stop()

    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


def run_sequence(scf, sequence):
    try:
        cf = scf.cf

        take_off(cf, sequence[0])
        for position in sequence:
            print('Setting position {}'.format(position))
            end_time = time.time() + position[3]
            while time.time() < end_time:
                cf.commander.send_position_setpoint(position[0],
                                                    position[1],
                                                    position[2], 0)
                time.sleep(0.1)
        land(cf, sequence[-1])
    except Exception as e:
        print(e)


if __name__ == '__main__':
    # logging.basicConfig(level=logging.DEBUG)
    cflib.crtp.init_drivers()

    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        # If the copters are started in their correct positions this is
        # probably not needed. The Kalman filter will have time to converge
        # any way since it takes a while to start them all up and connect. We
        # keep the code here to illustrate how to do it.
        # swarm.reset_estimators()

        # The current values of all parameters are downloaded as a part of the
        # connections sequence. Since we have 10 copters this is clogging up
        # communication and we have to wait for it to finish before we start
        # flying.
        print('Waiting for parameters to be downloaded...')
        swarm.parallel(wait_for_param_download)

        swarm.parallel(arm)

        swarm.parallel(run_sequence, args_dict=seq_args)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the first Crazyflie found, looks for
EEPROM memories and writes the default values in it.
"""
import logging
import sys
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import MemoryElement
from cflib.utils import uri_helper

uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class EEPROMExample:
    """
    Simple example listing the EEPROMs found and writes the default values
    in it.
    """

    def __init__(self, link_uri):
        """ Initialize and run the example with the specified link_uri """

        # Create a Crazyflie object without specifying any cache dirs
        self._cf = Crazyflie()

        # Connect some callbacks from the Crazyflie API
        self._cf.connected.add_callback(self._connected)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)

        print('Connecting to %s' % link_uri)

        # Try to connect to the Crazyflie
        self._cf.open_link(link_uri)

        # Variable used to keep main loop occupied until disconnect
        self.is_connected = True
        self.should_disconnect = False

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""
        print('Connected to %s' % link_uri)

        mems = self._cf.mem.get_mems(MemoryElement.TYPE_I2C)
        print('Found {} EEPOM(s)'.format(len(mems)))
        if len(mems) > 0:
            print('Writing default configuration to'
                  ' memory {}'.format(mems[0].id))

            elems = mems[0].elements
            elems['version'] = 1
            elems['pitch_trim'] = 0.0
            elems['roll_trim'] = 0.0
            elems['radio_channel'] = 80
            elems['radio_speed'] = 2
            elems['radio_address'] = 0xE7E7E7E7E7

            mems[0].write_data(self._data_written)

    def _data_written(self, mem, addr):
        print('Data written, reading back...')
        mem.update(self._data_updated)

    def _data_updated(self, mem):
        print('Updated id={}'.format(mem.id))
        print('\tType      : {}'.format(mem.type))
        print('\tSize      : {}'.format(mem.size))
        print('\tValid     : {}'.format(mem.valid))
        print('\tElements  : ')
        for key in mem.elements:
            print('\t\t{}={}'.format(key, mem.elements[key]))

        self.should_disconnect = True

    def _stab_log_error(self, logconf, msg):
        """Callback from the log API when an error occurs"""
        print('Error when logging %s: %s' % (logconf.name, msg))

    def _stab_log_data(self, timestamp, data, logconf):
        """Callback from the log API when data arrives"""
        print('[%d][%s]: %s' % (timestamp, logconf.name, data))

    def _connection_failed(self, link_uri, msg):
        """Callback when connection initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))
        self.is_connected = False

    def _connection_lost(self, link_uri, msg):
        """Callback when disconnected after a connection has been made (i.e
        Crazyflie moves out of range)"""
        print('Connection to %s lost: %s' % (link_uri, msg))

    def _disconnected(self, link_uri):
        """Callback when the Crazyflie is disconnected (called in all cases)"""
        print('Disconnected from %s' % link_uri)
        self.is_connected = False


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    le = EEPROMExample(uri)

    # The Crazyflie lib doesn't contain anything to keep the application alive,
    # so this is where your application should do something. In our case we
    # are just waiting until we are disconnected.
    try:
        while le.is_connected:
            if le.should_disconnect:
                le._cf.close_link()
            time.sleep(1)
    except KeyboardInterrupt:
        sys.exit(1)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the crazyflie at `URI` and runs a figure 8
sequence. This script requires some kind of location system, it has been
tested with the flow deck and the lighthouse positioning system.

Change the URI variable to your Crazyflie configuration.
"""
import logging
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf

        reset_estimator(scf)
        time.sleep(1)

        # Arm the Crazyflie
        cf.platform.send_arming_request(True)
        time.sleep(1.0)

        for y in range(10):
            cf.commander.send_hover_setpoint(0, 0, 0, y / 25)
            time.sleep(0.1)

        for _ in range(20):
            cf.commander.send_hover_setpoint(0, 0, 0, 0.4)
            time.sleep(0.1)

        for _ in range(50):
            cf.commander.send_hover_setpoint(0.5, 0, 36 * 2, 0.4)
            time.sleep(0.1)

        for _ in range(50):
            cf.commander.send_hover_setpoint(0.5, 0, -36 * 2, 0.4)
            time.sleep(0.1)

        for _ in range(20):
            cf.commander.send_hover_setpoint(0, 0, 0, 0.4)
            time.sleep(0.1)

        for y in range(10):
            cf.commander.send_hover_setpoint(0, 0, 0, (10 - y) / 25)
            time.sleep(0.1)

        cf.commander.send_stop_setpoint()
        # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
        cf.commander.send_notify_setpoint_stop()



import socket
import struct
from threading import Lock

from . import CPXPacket

found_serial = True
try:
    import serial
except ImportError:
    found_serial = False


class CPXTransport:
    def __init__(self):
        raise NotImplementedError('Cannot be used')

    # Change this to URI?
    def connect(host, port):
        raise NotImplementedError('Cannot be used')

    def disconnect():
        raise NotImplementedError('Cannot be used')

    def send(self, data):
        raise NotImplementedError('Cannot be used')

    def receive(self, size):
        raise NotImplementedError('Cannot be used')


class SocketTransport(CPXTransport):
    def __init__(self, host, port):
        print('CPX socket transport')
        self._host = host
        self._port = port

        self.connect()

    def connect(self):
        print('Connecting to socket on {}:{}...'.format(self._host, self._port))
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self._host, self._port))
        print('Connected')

    def disconnect(self):
        print('Closing transport')
        self._socket.shutdown(socket.SHUT_WR)
        self._socket.close()
        self._socket = None

    def writePacket(self, packet):
        data = bytearray(struct.pack('H', packet.length+2))
        data += packet.wireData
        self._socket.send(data)

    def _readData(self, size):
        data = bytearray()
        while len(data) < size and self._socket is not None:
            data.extend(self._socket.recv(size-len(data)))
        return data

    def readPacket(self):
        size = struct.unpack('H', self._readData(2))[0]

        data = self._readData(size)

        packet = CPXPacket()
        packet.wireData = data

        return packet

        def __del__(self):
            print('Socket transport is being destroyed!')


class UARTTransport(CPXTransport):
    def __init__(self, device, baudrate):
        print('CPX UART transport')
        self._device = device
        self._baudrate = baudrate
        self._serial = None
        self._cts = False
        self._lock = Lock()

        self.connect()

    def connect(self):
        print('Connecting to UART on {} @ {}...'.format(self._device, self._baudrate))
        self._serial = serial.Serial(self._device, self._baudrate, timeout=None)

        isInSync = False

        while not isInSync:
            start = self._serial.read(1)[0]
            print(start)
            if start == 0xFF:
                print('Got start')
                size = self._serial.read(1)[0]
                print(size)
                if size == 0x00:
                    isInSync = True
                    self.cts = True

        # Send back sync
        self._serial.write([0xFF, 0x00])

        print('Connected')

    def _calcXORchecksum(self, data):
        checksum = 0
        for i in data:
            checksum ^= i
        return checksum

    def disconnect(self):
        print('Closing transport')
        self._serial.close()
        self._serial = None

    def writePacket(self, packet):
        self._lock.acquire()
        data = packet.wireData
        if len(data) > 100:
            raise 'Packet too large!'

        buff = bytearray([0xFF, len(data)])
        buff.extend(data)
        buff.extend([self._calcXORchecksum(buff)])
        self._serial.write(buff)

    def readPacket(self):
        size = 0
        while size == 0:
            start = self._serial.read(1)[0]
            if start == 0xFF:
                size = self._serial.read(1)[0]
                if size == 0:
                    self._lock.release()
                else:
                    data = self._serial.read(size)  # Size is excluding start (0xFF) and checksum at end
                    crc = self._serial.read(1)
                    # CRC includes start and size
                    calculated_crc = self._calcXORchecksum(bytes([start, size]) + data)
                    if calculated_crc != ord(crc):
                        print('CRC error!')
                    # Send CTS
                    self._serial.write([0xFF, 0x00])

                    packet = CPXPacket()
                    packet.wireData = data

        return packet


class CRTPTransport(CPXTransport):
    def __init__(self):
        print('CPX CRTP transport')

    # This connection will not really work...
    def connect(host, port):
        pass

    def disconnect():
        pass

    def send(self, data):
        pass

    def receive(self, size):
        pass



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import NamedTuple

import numpy as np
import numpy.typing as npt

from .ippe_cf import IppeCf
from cflib.localization.lighthouse_types import LhBsCfPoses
from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import LhException
from cflib.localization.lighthouse_types import Pose


ArrayFloat = npt.NDArray[np.float64]


class BsPairIds(NamedTuple):
    """A type representing the ids of a pair of base stations"""
    bs1: int
    bs2: int


class BsPairPoses(NamedTuple):
    """A type representing the poses of a pair of base stations"""
    bs1: Pose
    bs2: Pose


class LighthouseInitialEstimator:
    """
    Make initial estimates of base station and CF poses using IPPE (analytical solution).
    The estimates are not good enough to use for flight but is a starting point for further
    calculations.
    """

    OUTLIER_DETECTION_ERROR = 0.5

    @classmethod
    def estimate(cls, matched_samples: list[LhCfPoseSample], sensor_positions: ArrayFloat) -> tuple[
            LhBsCfPoses, list[LhCfPoseSample]]:
        """
        Make a rough estimate of the poses of all base stations and CF poses found in the samples.

        The pose of the Crazyflie in the first sample is used as a reference and will define the
        global reference frame.

        :param matched_samples: A list of samples with lighthouse angles.
        :param sensor_positions: An array with the sensor positions on the lighthouse deck (3D, CF ref frame)
        :return: an estimate of base station and Crazyflie poses, as well as a cleaned version of matched_samples where
                 outliers are removed.
        """

        bs_positions = cls._find_solutions(matched_samples, sensor_positions)
        # bs_positions is a map from bs-id-pair to position, where the position is the position of the second
        # bs, as seen from the first bs (in the first bs ref frame).

        bs_poses_ref_cfs, cleaned_matched_samples = cls._angles_to_poses(
            matched_samples, sensor_positions, bs_positions)

        # Use the first CF pose as the global reference frame. The pose of the first base station (as estimated by ippe)
        # is used as the "true" position (reference)
        reference_bs_pose = None
        for bs_pose_ref_cfs in bs_poses_ref_cfs:
            if len(bs_pose_ref_cfs) > 0:
                bs_id, reference_bs_pose = list(bs_pose_ref_cfs.items())[0]
                break

        if reference_bs_pose is None:
            raise LhException('Too little data, no reference')
        bs_poses: dict[int, Pose] = {bs_id: reference_bs_pose}

        # Calculate the pose of the remaining base stations, based on the pose of the first CF
        cls._estimate_remaining_bs_poses(bs_poses_ref_cfs, bs_poses)

        # Now that we have estimated the base station poses, estimate the poses of the CF in all the samples
        cf_poses = cls._estimate_cf_poses(bs_poses_ref_cfs, bs_poses)

        return LhBsCfPoses(bs_poses, cf_poses), cleaned_matched_samples

    @classmethod
    def _find_solutions(cls, matched_samples: list[LhCfPoseSample], sensor_positions: ArrayFloat
                        ) -> dict[BsPairIds, ArrayFloat]:
        """
        Find the pose of all base stations, in the reference frame of other base stations.

        Ippe finds two mirror solutions for each sample and the bulk of this method is geared towards finding the
        correct one. The outline of the process is:
        1. For each sample collect all possible permutations (4) of the base station positions
        2. Aggregate the possible positions of all samples in clusters. This is done per base station pair.
        3. Pick the "best" cluster as the correct permutation. The idea is that the mirror solutions will be spread
           out in space, while the correct one will end up more or less in the same spot for all samples.

        :param matched_samples: List of matched samples
        :param sensor_positions: list of sensor positions on the lighthouse deck, CF reference frame
        :return: Base stations poses in the reference frame of the other base stations. The data is organized as a
                 dictionary of tuples with base station id pairs, mapped to positions. For instance the entry with key
                 (2, 1) contains the position of base station 1, in the base station 2 reference frame.
        """

        position_permutations: dict[BsPairIds, list[list[ArrayFloat]]] = {}
        for sample in matched_samples:
            solutions: dict[int, BsPairPoses] = {}
            for bs, angles in sample.angles_calibrated.items():
                projections = angles.projection_pair_list()
                estimates_ref_bs = IppeCf.solve(sensor_positions, projections)
                estimates_ref_cf = cls._convert_estimates_to_cf_reference_frame(estimates_ref_bs)
                solutions[bs] = estimates_ref_cf

            cls._add_solution_permutations(solutions, position_permutations)

        return cls._find_most_likely_positions(position_permutations)

    @classmethod
    def _add_solution_permutations(cls, solutions: dict[int, BsPairPoses],
                                   position_permutations: dict[BsPairIds, list[list[ArrayFloat]]]):
        """
        Add the possible permutations of base station positions for a sample to a collection of aggregated positions.
        The aggregated collection contains base station positions in the reference frame of other base stations.

        :param solutions: All possible positions of the base stations, in the reference frame of the Crazyflie in one
                          sample
        :param position_permutations: Aggregated possible solutions. A dictionary with base station pairs as keys,
                                      mapped to lists of lists of possible positions. For instance, the entry for (2, 1)
                                      would contain a list of lists with 4 positions each, for where base station 1
                                      might be located in the base station 2 reference frame.
        """
        ids = sorted(solutions.keys())

        for i, id_i in enumerate(ids):
            solution_i = solutions[id_i]

            for j in range(i + 1, len(ids)):
                id_j = ids[j]

                solution_j = solutions[id_j]

                pose1 = solution_i[0].inv_rotate_translate_pose(solution_j[0])
                pose2 = solution_i[0].inv_rotate_translate_pose(solution_j[1])
                pose3 = solution_i[1].inv_rotate_translate_pose(solution_j[0])
                pose4 = solution_i[1].inv_rotate_translate_pose(solution_j[1])

                pair = BsPairIds(id_i, id_j)
                if pair not in position_permutations:
                    position_permutations[pair] = []
                position_permutations[pair].append([pose1.translation, pose2.translation,
                                                   pose3.translation, pose4.translation])

    @classmethod
    def _angles_to_poses(cls, matched_samples: list[LhCfPoseSample], sensor_positions: ArrayFloat,
                         bs_positions: dict[BsPairIds, ArrayFloat]) -> tuple[list[dict[int, Pose]],
                                                                             list[LhCfPoseSample]]:
        """
        Estimate the base station poses in the Crazyflie reference frames, for each sample.

        Use Ippe again to find the possible poses of the bases stations and pick the one that best matches the position
        in bs_positions.

        :param matched_samples: List of samples
        :param sensor_positions: Positions of the sensors on the lighthouse deck (CF ref frame)
        :param bs_positions: Dictionary of base station positions (other base station ref frame)
        :return: A list of dictionaries from base station to Pose of all base stations, for each sample, as well as
                 a version of the matched_samples where outliers are removed
        """
        result: list[dict[int, Pose]] = []

        cleaned_matched_samples: list[LhCfPoseSample] = []

        for sample in matched_samples:
            solutions: dict[int, BsPairPoses] = {}
            for bs, angles in sample.angles_calibrated.items():
                projections = angles.projection_pair_list()
                estimates_ref_bs = IppeCf.solve(sensor_positions, projections)
                estimates_ref_cf = cls._convert_estimates_to_cf_reference_frame(estimates_ref_bs)
                solutions[bs] = estimates_ref_cf

            poses: dict[int, Pose] = {}
            ids = sorted(solutions.keys())
            first = ids[0]
            is_sample_valid = True

            for other in ids[1:]:
                pair_ids = BsPairIds(first, other)
                expected = bs_positions[pair_ids]

                success, pair_poses = cls._choose_solutions(solutions[first], solutions[other], expected)
                if success:
                    poses[pair_ids.bs1] = pair_poses.bs1
                    poses[pair_ids.bs2] = pair_poses.bs2
                else:
                    is_sample_valid = False
                    break

            if is_sample_valid:
                result.append(poses)
                cleaned_matched_samples.append(sample)

        return result, cleaned_matched_samples

    @classmethod
    def _choose_solutions(cls, solutions_1: BsPairPoses, solutions_2: BsPairPoses,
                          expected: ArrayFloat) -> tuple[bool, BsPairPoses]:
        """Pick the base pose solutions for a pair of base stations, based on the position in expected"""

        min_dist = 100000.0
        best = BsPairPoses(Pose(), Pose())
        success = True

        for solution_1 in solutions_1:
            for solution_2 in solutions_2:
                pose_second_bs_ref_fr_first = solution_1.inv_rotate_translate_pose(solution_2)
                dist = np.linalg.norm(expected - pose_second_bs_ref_fr_first.translation)
                if dist < min_dist:
                    min_dist = dist
                    best = BsPairPoses(solution_1, solution_2)

        if min_dist > cls.OUTLIER_DETECTION_ERROR:
            success = False

        return success, best

    @classmethod
    def _find_most_likely_positions(cls, position_permutations: dict[BsPairIds,
                                    list[list[ArrayFloat]]]) -> dict[BsPairIds, ArrayFloat]:
        """
        Find the most likely base station positions from all the possible permutations.

        Sort the permutations into buckets based on how close they are to the solutions in the first sample. Solutions
        that are "too" far away and discarded. The bucket with the most samples in, is considered the best.
        """
        result: dict[BsPairIds, ArrayFloat] = {}

        for pair, position_lists in position_permutations.items():
            # Use first as reference to map the others to
            bucket_ref_positions = position_lists[0]
            buckets: list[list[ArrayFloat]] = [[], [], [], []]

            cls._map_positions_to_ref(bucket_ref_positions, position_lists, buckets)
            best_pos = cls._find_best_position_bucket(buckets)
            result[pair] = best_pos

        return result

    @classmethod
    def _map_positions_to_ref(cls, bucket_ref_positions: list[ArrayFloat],
                              position_lists: list[list[ArrayFloat]],
                              buckets: list[list[ArrayFloat]]) -> None:
        """
        Sort solution into buckets based on the distance to the reference position. If no bucket is close enough,
        the solution is discarded.
        """

        accept_radius = 0.8

        for pos_list in position_lists:
            for pos in pos_list:
                for i, ref in enumerate(bucket_ref_positions):
                    if np.linalg.norm(pos - ref) < accept_radius:
                        buckets[i].append(pos)
                        break

    @classmethod
    def _find_best_position_bucket(cls, buckets: list[list[ArrayFloat]]) -> ArrayFloat:
        """
        Find the bucket with the most solutions in, this is considered to be the correct solution.
        The final result is the mean of the solution in the bucket.
        """
        max_len = 0
        max_i = 0
        for i, bucket in enumerate(buckets):
            if len(bucket) > max_len:
                max_len = len(bucket)
                max_i = i

        pos = np.mean(buckets[max_i], axis=0)
        return pos

    @classmethod
    def _convert_estimates_to_cf_reference_frame(cls, estimates_ref_bs: list[IppeCf.Solution]) -> BsPairPoses:
        """
        Convert the two ippe solutions from the base station reference frame to the CF reference frame
        """
        rot_1 = estimates_ref_bs[0].R.transpose()
        t_1 = np.dot(rot_1, -estimates_ref_bs[0].t)

        rot_2 = estimates_ref_bs[1].R.transpose()
        t_2 = np.dot(rot_2, -estimates_ref_bs[1].t)

        return BsPairPoses(Pose(rot_1, t_1), Pose(rot_2, t_2))

    @classmethod
    def _estimate_remaining_bs_poses(cls, bs_poses_ref_cfs: list[dict[int, Pose]], bs_poses: dict[int, Pose]) -> None:
        """
        Based on one base station pose, estimate the other base station poses.

        The process is iterative and runs until all poses are found. Assume we know the pose of base station 0, and we
        have information of base station pairs (0, 2) and (2, 3), from this we can first derive the pose of 2 and after
        that the pose of 3.
        """
        # Find all base stations in the list
        all_bs = set()
        for initial_est_bs_poses in bs_poses_ref_cfs:
            all_bs.update(initial_est_bs_poses.keys())

        # Remove the reference base stations that we already have the poses for
        to_find = all_bs - bs_poses.keys()

        # run through the list of samples until we manage to find them all
        remaining = len(to_find)
        while remaining > 0:
            buckets: dict[int, list[Pose]] = {}
            for bs_poses_in_sample in bs_poses_ref_cfs:
                unknown = to_find.intersection(bs_poses_in_sample.keys())
                known = set(bs_poses.keys()).intersection(bs_poses_in_sample.keys())

                # We need (at least) one known bs pose to use when transforming the other poses to the global ref frame
                if len(known) > 0:
                    known_bs = list(known)[0]

                    # The known BS pose in the global reference frame
                    known_global = bs_poses[known_bs]
                    # The known BS pose in the CF reference frame (of this sample)
                    known_cf = bs_poses_in_sample[known_bs]

                    for bs_id in unknown:
                        # The unknown BS pose in the CF reference frame (of this sample)
                        unknown_cf = bs_poses_in_sample[bs_id]
                        # Finally we can calculate the BS pose in the global reference frame
                        bs_pose = cls._map_pose_to_ref_frame(known_global, known_cf, unknown_cf)
                        if bs_id not in buckets:
                            buckets[bs_id] = []
                        buckets[bs_id].append(bs_pose)

            # Average over poses and add to bs_poses
            for bs_id, poses in buckets.items():
                bs_poses[bs_id] = cls._avarage_poses(poses)

            to_find = all_bs - bs_poses.keys()
            if len(to_find) == 0:
                break

            if len(to_find) == remaining:
                raise LhException('Can not link positions between all base stations')

            remaining = len(to_find)

    @classmethod
    def _avarage_poses(cls, poses: list[Pose]) -> Pose:
        """
        Averaging of quaternions to get the "average" orientation of multiple samples.
        From https://stackoverflow.com/a/61013769
        """
        def q_average(Q, W=None):
            if W is not None:
                Q *= W[:, None]
            eigvals, eigvecs = np.linalg.eig(Q.T@Q)
            return eigvecs[:, eigvals.argmax()]

        positions = map(lambda x: x.translation, poses)
        average_pos = np.average(np.array(list(positions)), axis=0)

        quats = map(lambda x: x.rot_quat, poses)
        average_quaternion = q_average(np.array(list(quats)))

        return Pose.from_quat(R_quat=average_quaternion, t_vec=average_pos)

    @classmethod
    def _estimate_cf_poses(cls, bs_poses_ref_cfs: list[dict[int, Pose]], bs_poses: dict[int, Pose]) -> list[Pose]:
        """
        Find the pose of the Crazyflie in all samples, based on the base station poses.
        """
        cf_poses: list[Pose] = []

        for est_ref_cf in bs_poses_ref_cfs:
            # Average the global pose based on all base stations
            poses = []
            for bs_id, pose_cf in est_ref_cf.items():
                pose_global = bs_poses[bs_id]
                est_ref_global = cls._map_cf_pos_to_cf_pos(pose_global, pose_cf)
                poses.append(est_ref_global)

            cf_poses.append(cls._avarage_poses(poses))

        return cf_poses

    @classmethod
    def _map_pose_to_ref_frame(cls, pose1_ref1: Pose, pose1_ref2: Pose, pose2_ref2: Pose) -> Pose:
        """
        Express pose2 in reference system 1, given pose1 in both reference system 1 and 2
        """
        R_o2_in_1, t_o2_in_1 = cls._map_cf_pos_to_cf_pos(pose1_ref1, pose1_ref2).matrix_vec

        t = np.dot(R_o2_in_1, pose2_ref2.translation) + t_o2_in_1
        R = np.dot(R_o2_in_1, pose2_ref2.rot_matrix)

        return Pose(R, t)

    @classmethod
    def _map_cf_pos_to_cf_pos(cls, pose1_ref1: Pose, pose1_ref2: Pose) -> Pose:
        """
        Find the rotation/translation from ref1 to ref2 given a pose,
        that is the returned Pose will tell us where the origin in ref2 is,
        expressed in ref1
        """

        R_inv_ref2 = np.matrix.transpose(pose1_ref2.rot_matrix)
        R = np.dot(pose1_ref1.rot_matrix, R_inv_ref2)
        t = pose1_ref1.translation - np.dot(R, pose1_ref2.translation)

        return Pose(R, t)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie


class Multiranger:
    FRONT = 'range.front'
    BACK = 'range.back'
    LEFT = 'range.left'
    RIGHT = 'range.right'
    UP = 'range.up'
    DOWN = 'range.zrange'

    def __init__(self, crazyflie, rate_ms=100, zranger=False):
        if isinstance(crazyflie, SyncCrazyflie):
            self._cf = crazyflie.cf
        else:
            self._cf = crazyflie
        self._log_config = self._create_log_config(rate_ms)

        self._up_distance = None
        self._front_distance = None
        self._back_distance = None
        self._left_distance = None
        self._right_distance = None
        self._down_distance = None

    def _create_log_config(self, rate_ms):
        log_config = LogConfig('multiranger', rate_ms)
        log_config.add_variable(self.FRONT)
        log_config.add_variable(self.BACK)
        log_config.add_variable(self.LEFT)
        log_config.add_variable(self.RIGHT)
        log_config.add_variable(self.UP)
        log_config.add_variable(self.DOWN)

        log_config.data_received_cb.add_callback(self._data_received)

        return log_config

    def start(self):
        self._cf.log.add_config(self._log_config)
        self._log_config.start()

    def _convert_log_to_distance(self, data):
        if data >= 8000:
            return None
        else:
            return data / 1000.0

    def _data_received(self, timestamp, data, logconf):
        self._up_distance = self._convert_log_to_distance(data[self.UP])
        self._front_distance = self._convert_log_to_distance(data[self.FRONT])
        self._back_distance = self._convert_log_to_distance(data[self.BACK])
        self._left_distance = self._convert_log_to_distance(data[self.LEFT])
        self._right_distance = self._convert_log_to_distance(data[self.RIGHT])
        if self.DOWN in data:
            self._down_distance = self._convert_log_to_distance(data[self.DOWN]
                                                                )

    def stop(self):
        self._log_config.delete()

    @property
    def up(self):
        return self._up_distance

    @property
    def left(self):
        return self._left_distance

    @property
    def right(self):
        return self._right_distance

    @property
    def front(self):
        return self._front_distance

    @property
    def back(self):
        return self._back_distance

    @property
    def down(self):
        return self._down_distance

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example of how to write to the Lighthouse base station geometry
and calibration memory in a Crazyflie
"""
import logging
from threading import Event

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import LighthouseBsCalibration
from cflib.crazyflie.mem import LighthouseBsGeometry
from cflib.crazyflie.mem import LighthouseMemHelper
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class WriteMem:
    def __init__(self, uri, geo_dict, calib_dict):
        self._event = Event()

        with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
            helper = LighthouseMemHelper(scf.cf)

            helper.write_geos(geo_dict, self._data_written)
            self._event.wait()

            self._event.clear()

            helper.write_calibs(calib_dict, self._data_written)
            self._event.wait()

    def _data_written(self, success):
        if success:
            print('Data written')
        else:
            print('Write failed')

        self._event.set()


if __name__ == '__main__':
    # URI to the Crazyflie to connect to
    uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    bs1geo = LighthouseBsGeometry()
    bs1geo.origin = [1.0, 2.0, 3.0]
    bs1geo.rotation_matrix = [
        [4.0, 5.0, 6.0],
        [7.0, 8.0, 9.0],
        [10.0, 11.0, 12.0],
    ]
    bs1geo.valid = True

    bs2geo = LighthouseBsGeometry()
    bs2geo.origin = [21.0, 22.0, 23.0]
    bs2geo.rotation_matrix = [
        [24.0, 25.0, 26.0],
        [27.0, 28.0, 29.0],
        [30.0, 31.0, 32.0],
    ]
    bs2geo.valid = True

    bs1calib = LighthouseBsCalibration()
    bs1calib.sweeps[0].phase = 1.0
    bs1calib.sweeps[0].tilt = 2.0
    bs1calib.sweeps[0].curve = 3.0
    bs1calib.sweeps[0].gibmag = 4.0
    bs1calib.sweeps[0].gibphase = 5.0
    bs1calib.sweeps[0].ogeemag = 6.0
    bs1calib.sweeps[0].ogeephase = 7.0
    bs1calib.sweeps[1].phase = 1.1
    bs1calib.sweeps[1].tilt = 2.1
    bs1calib.sweeps[1].curve = 3.1
    bs1calib.sweeps[1].gibmag = 4.1
    bs1calib.sweeps[1].gibphase = 5.1
    bs1calib.sweeps[1].ogeemag = 6.1
    bs1calib.sweeps[1].ogeephase = 7.1
    bs1calib.uid = 1234
    bs1calib.valid = True

    bs2calib = LighthouseBsCalibration()
    bs2calib.sweeps[0].phase = 1.5
    bs2calib.sweeps[0].tilt = 2.5
    bs2calib.sweeps[0].curve = 3.5
    bs2calib.sweeps[0].gibmag = 4.5
    bs2calib.sweeps[0].gibphase = 5.5
    bs2calib.sweeps[0].ogeemag = 6.5
    bs2calib.sweeps[0].ogeephase = 7.5
    bs2calib.sweeps[1].phase = 1.51
    bs2calib.sweeps[1].tilt = 2.51
    bs2calib.sweeps[1].curve = 3.51
    bs2calib.sweeps[1].gibmag = 4.51
    bs2calib.sweeps[1].gibphase = 5.51
    bs2calib.sweeps[1].ogeemag = 6.51
    bs2calib.sweeps[1].ogeephase = 7.51
    bs2calib.uid = 9876
    bs2calib.valid = True

    # Note: base station ids (channels) are 0-indexed
    geo_dict = {0: bs1geo, 1: bs2geo}
    calib_dict = {0: bs1calib, 1: bs2calib}

    WriteMem(uri, geo_dict, calib_dict)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the first Crazyflie found, triggers
reading of all the parameters and displays their values. It then modifies
one parameter and reads back it's value. Finally it disconnects.
"""
import logging
import random
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.utils import uri_helper

address = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class ParamExample:
    """
    Simple logging example class that logs the Stabilizer from a supplied
    link uri and disconnects after 5s.
    """

    def __init__(self, link_uri):
        """ Initialize and run the example with the specified link_uri """

        self._cf = Crazyflie(rw_cache='./cache')

        # Connect some callbacks from the Crazyflie API
        self._cf.connected.add_callback(self._connected)
        self._cf.fully_connected.add_callback(self._fully_connected)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)

        print('Connecting to %s' % link_uri)

        # Try to connect to the Crazyflie
        self._cf.open_link(link_uri)

        # Variable used to keep main loop occupied until disconnect
        self.is_connected = True
        self.should_disconnect = False

        self._param_check_list = []
        self._param_groups = []

        random.seed()

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded. Parameter values are not downloaded yet."""
        print('Connected to %s' % link_uri)

        # Print the param TOC
        p_toc = self._cf.param.toc.toc
        for group in sorted(p_toc.keys()):
            print('{}'.format(group))
            for param in sorted(p_toc[group].keys()):
                print('\t{}'.format(param))
                self._param_check_list.append('{0}.{1}'.format(group, param))
            self._param_groups.append('{}'.format(group))
            # For every group, register the callback
            self._cf.param.add_update_callback(group=group, name=None, cb=self._param_callback)

        # You can also register a callback for a specific group.name combo
        self._cf.param.add_update_callback(group='cpu', name='flash',
                                           cb=self._cpu_flash_callback)

        print('')

    def _fully_connected(self, link_uri):
        """This callback is called when the Crazyflie has been connected and all parameters have been
        downloaded. It is now OK to set and get parameters."""
        print(f'Parameters downloaded to {link_uri}')

        # We can get a parameter value directly without using a callback
        value = self._cf.param.get_value('pid_attitude.pitch_kd')
        print(f'Value read with get() is {value}')

        # When a parameter is set, the callback is called with the new value
        self._cf.param.add_update_callback(group='pid_attitude', name='pitch_kd', cb=self._a_pitch_kd_callback)
        # When setting a value the parameter is automatically read back
        # and the registered callbacks will get the updated value
        self._cf.param.set_value('pid_attitude.pitch_kd', 0.1234)

    def _cpu_flash_callback(self, name, value):
        """Specific callback for the cpu.flash parameter"""
        print('The connected Crazyflie has {}kb of flash'.format(value))

    def _param_callback(self, name, value):
        """Generic callback registered for all the groups"""
        print('{0}: {1}'.format(name, value))

        # Remove each parameter from the list when fetched
        self._param_check_list.remove(name)
        if len(self._param_check_list) == 0:
            print('Have fetched all parameter values.')

            # Remove all the group callbacks
            for g in self._param_groups:
                self._cf.param.remove_update_callback(group=g, cb=self._param_callback)

    def _a_pitch_kd_callback(self, name, value):
        """Callback for pid_attitude.pitch_kd"""
        print('Read back: {0}={1}'.format(name, value))

        # This is the end of the example, signal to close link
        self.should_disconnect = True

    def _connection_failed(self, link_uri, msg):
        """Callback when connection initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))
        self.is_connected = False

    def _connection_lost(self, link_uri, msg):
        """Callback when disconnected after a connection has been made (i.e
        Crazyflie moves out of range)"""
        print('Connection to %s lost: %s' % (link_uri, msg))

    def _disconnected(self, link_uri):
        """Callback when the Crazyflie is disconnected (called in all cases)"""
        print('Disconnected from %s' % link_uri)
        self.is_connected = False


if __name__ == '__main__':
    import logging
    logging.getLogger().setLevel(logging.INFO)
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()
    pe = ParamExample(address)

    # The Crazyflie lib doesn't contain anything to keep the application
    # alive, so this is where your application should do something. In our
    # case we are just waiting until we are disconnected.
    while pe.is_connected:
        if pe.should_disconnect:
            pe._cf.close_link()
        time.sleep(1)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Enables reading/writing of parameter values to/from the Crazyflie.

When a Crazyflie is connected it's possible to download a TableOfContent of all
the parameters that can be written/read.

"""
import copy
import errno
import logging
import struct
from collections import namedtuple
from queue import Empty
from queue import Queue
from threading import Event
from threading import Lock
from threading import Thread

from .toc import Toc
from .toc import TocFetcher
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['Param', 'ParamTocElement']

logger = logging.getLogger(__name__)

# Possible states
IDLE = 0
WAIT_TOC = 1
WAIT_READ = 2
WAIT_WRITE = 3

TOC_CHANNEL = 0
READ_CHANNEL = 1
WRITE_CHANNEL = 2
MISC_CHANNEL = 3

MISC_SETBYNAME = 0
MISC_VALUE_UPDATED = 1
MISC_GET_EXTENDED_TYPE = 2
MISC_PERSISTENT_STORE = 3
MISC_PERSISTENT_GET_STATE = 4
MISC_PERSISTENT_CLEAR = 5
MISC_GET_DEFAULT_VALUE = 6

PersistentParamState = namedtuple('PersistentParamState', 'is_stored default_value stored_value')


# One element entry in the TOC


class ParamTocElement:
    """An element in the Log TOC."""

    RW_ACCESS = 0
    RO_ACCESS = 1

    EXTENDED_PERSISTENT = 1

    types = {0x08: ('uint8_t', '<B'),
             0x09: ('uint16_t', '<H'),
             0x0A: ('uint32_t', '<L'),
             0x0B: ('uint64_t', '<Q'),
             0x00: ('int8_t', '<b'),
             0x01: ('int16_t', '<h'),
             0x02: ('int32_t', '<i'),
             0x03: ('int64_t', '<q'),
             0x05: ('FP16', ''),
             0x06: ('float', '<f'),
             0x07: ('double', '<d')}

    def __init__(self, ident=0, data=None):
        """TocElement creator. Data is the binary payload of the element."""
        self.ident = ident
        self.persistent = False
        self.extended = False
        if (data):
            strs = struct.unpack('s' * len(data[1:]), data[1:])
            s = ''
            for ch in strs:
                s += ch.decode('ISO-8859-1')
            strs = s.split('\x00')
            self.group = strs[0]
            self.name = strs[1]

            metadata = data[0]
            if isinstance(metadata, str):
                metadata = ord(metadata)

            # If the fouth byte (1 << 4) (0x10) is set we have extended
            # type information for this element.
            self.extended = ((metadata & 0x10) != 0)

            self.ctype = self.types[metadata & 0x0F][0]
            self.pytype = self.types[metadata & 0x0F][1]
            if ((metadata & 0x40) != 0):
                self.access = ParamTocElement.RO_ACCESS
            else:
                self.access = ParamTocElement.RW_ACCESS

    def get_readable_access(self):
        if (self.access == ParamTocElement.RO_ACCESS):
            return 'RO'
        return 'RW'

    def is_extended(self):
        return self.extended

    def mark_persistent(self):
        self.persistent = True

    def is_persistent(self):
        return self.persistent


class Param():
    """
    Used to read and write parameter values in the Crazyflie.
    """

    def __init__(self, crazyflie):
        self.toc = Toc()

        self.cf = crazyflie
        self._useV2 = False
        self.param_update_callbacks = {}
        self.group_update_callbacks = {}
        self.all_update_callback = Caller()
        self.param_updater = None

        self.cf.disconnected.add_callback(self._disconnected)
        self.cf.connection_requested.add_callback(self._connection_requested)

        self.all_updated = Caller()
        self.is_updated = False
        self._initialized = Event()

        self.values = {}

    def request_update_of_all_params(self):
        """Request an update of all the parameters in the TOC"""
        for group in self.toc.toc:
            for name in self.toc.toc[group]:
                complete_name = '%s.%s' % (group, name)
                self.request_param_update(complete_name)

    def _check_if_all_updated(self):
        """Check if all parameters from the TOC has at least been fetched
        once"""
        for g in self.toc.toc:
            if g not in self.values:
                return False
            for n in self.toc.toc[g]:
                if n not in self.values[g]:
                    return False

        return True

    def _param_updated(self, pk):
        """Callback with data for an updated parameter"""

        # This method handles both param value packets as well as misc param updated packets
        # The Misc packets have a command byte first and the variable id is shifted one byte
        # Misc packets are not supported for V1
        if pk.channel == MISC_CHANNEL:
            id_index = 1
        else:
            id_index = 0

        if self._useV2:
            var_id = struct.unpack('<H', pk.data[id_index:id_index + 2])[0]
        else:
            var_id = pk.data[0]
        element = self.toc.get_element_by_id(var_id)
        if element:
            if self._useV2:
                value = struct.unpack(element.pytype, pk.data[id_index + 2:])[0]
            else:
                value = struct.unpack(element.pytype, pk.data[1:])[0]
            value_s = value.__str__()
            complete_name = '%s.%s' % (element.group, element.name)

            # Save the value for synchronous access
            if element.group not in self.values:
                self.values[element.group] = {}
            self.values[element.group][element.name] = value_s

            logger.debug('Updated parameter [%s]' % complete_name)
            if complete_name in self.param_update_callbacks:
                self.param_update_callbacks[complete_name].call(complete_name, value_s)
            if element.group in self.group_update_callbacks:
                self.group_update_callbacks[element.group].call(complete_name, value_s)
            self.all_update_callback.call(complete_name, value_s)

            # Once all the parameters are updated call the
            # callback for "everything updated"
            if self._check_if_all_updated() and not self.is_updated:
                self.is_updated = True
                self._initialized.set()
                self.all_updated.call()
        else:
            logger.debug('Variable id [%d] not found in TOC', var_id)

    def remove_update_callback(self, group, name=None, cb=None):
        """Remove the supplied callback for a group or a group.name"""
        if not cb:
            return

        if not name:
            if group in self.group_update_callbacks:
                self.group_update_callbacks[group].remove_callback(cb)
        else:
            paramname = '{}.{}'.format(group, name)
            if paramname in self.param_update_callbacks:
                self.param_update_callbacks[paramname].remove_callback(cb)

    def add_update_callback(self, group=None, name=None, cb=None):
        """
        Add a callback for a specific parameter name. This callback will be
        executed when a new value is read from the Crazyflie.
        """
        if not group and not name:
            self.all_update_callback.add_callback(cb)
        elif not name:
            if group not in self.group_update_callbacks:
                self.group_update_callbacks[group] = Caller()
            self.group_update_callbacks[group].add_callback(cb)
        else:
            paramname = '{}.{}'.format(group, name)
            if paramname not in self.param_update_callbacks:
                self.param_update_callbacks[paramname] = Caller()
            self.param_update_callbacks[paramname].add_callback(cb)

    def refresh_toc(self, refresh_done_callback, toc_cache):
        """
        Initiate a refresh of the parameter TOC.
        """
        def refresh_done():
            extended_elements = list()

            for group in self.toc.toc:
                for element in self.toc.toc[group].values():
                    if element.is_extended():
                        extended_elements.append(element)

            if len(extended_elements) > 0:
                extended_type_fetcher = _ExtendedTypeFetcher(self.cf, self.toc)
                extended_type_fetcher.start()
                extended_type_fetcher.set_callback(refresh_done_callback)
                extended_type_fetcher.request_extended_types(extended_elements)
            else:
                refresh_done_callback()

        self._useV2 = self.cf.platform.get_protocol_version() >= 4
        toc_fetcher = TocFetcher(self.cf, ParamTocElement,
                                 CRTPPort.PARAM, self.toc,
                                 refresh_done, toc_cache)
        toc_fetcher.start()

    def _connection_requested(self, uri):
        # Reset the internal state on connect to make sure we have a clean state
        self.param_updater = _ParamUpdater(self.cf, self._useV2, self._param_updated)
        self.param_updater.start()
        self.is_updated = False
        self.toc = Toc()
        self.values = {}
        self._initialized.clear()

    def _disconnected(self, uri):
        logger.info('Disconnected, cleaning up threads')
        """Disconnected callback from Crazyflie API"""
        if self.param_updater is not None:
            self.param_updater.close()
            self.param_updater = None

        # Do not clear self.is_updated here as we might get spurious parameter updates later

        # Clear all values from the previous Crazyflie
        self.toc = Toc()
        self.values = {}

    def request_param_update(self, complete_name):
        """
        Request an update of the value for the supplied parameter.
        """
        if self.param_updater is None:
            raise Exception('Param updater not initialized, did you call open_connection?')
        self.param_updater.request_param_update(
            self.toc.get_element_id(complete_name))

    def set_value_raw(self, complete_name, type, value):
        """
        Set a parameter value using the complete name and the type. Does not
        need to have received the TOC.
        """
        char_array = bytes(complete_name.replace('.', '\0') + '\0', 'utf-8')
        len_array = len(char_array)

        # This gives us the type for the struct.pack
        pytype = ParamTocElement.types[type][1][1]

        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)
        pk.data = struct.pack(f'<B{len_array}sB{pytype}', 0, char_array, type, value)

        # We will not get an update callback when using raw (MISC_CHANNEL)
        # so just send.
        self.cf.send_packet(pk)

    def set_value(self, complete_name, value):
        """
        Set the value for the supplied parameter.
        """
        if not self._initialized.is_set():
            if self.cf.is_called_by_incoming_handler_thread():
                raise Exception('Can not set parameter from callback until fully connected.')
            if not self._initialized.wait(timeout=60):
                raise Exception('Connection timed out')

        element = self.toc.get_element_by_complete_name(complete_name)

        if not element:
            logger.warning("Cannot set value for [%s], it's not in the TOC!",
                           complete_name)
            raise KeyError('{} not in param TOC'.format(complete_name))
        elif element.access == ParamTocElement.RO_ACCESS:
            logger.debug('[%s] is read only, no trying to set value',
                         complete_name)
            raise AttributeError('{} is read-only!'.format(complete_name))
        else:
            varid = element.ident
            pk = CRTPPacket()
            pk.set_header(CRTPPort.PARAM, WRITE_CHANNEL)
            if self._useV2:
                pk.data = struct.pack('<H', varid)
            else:
                pk.data = struct.pack('<B', varid)

            if element.pytype == '<f' or element.pytype == '<d':
                value_nr = float(value)
            else:
                value_nr = int(value)

            pk.data += struct.pack(element.pytype, value_nr)
            self.param_updater.request_param_setvalue(pk)

    def get_value(self, complete_name, timeout=60):
        """
        Read a value for the supplied parameter. This can block for a period
        of time if the parameter values have not been fetched yet.
        """
        if not self._initialized.is_set():
            if self.cf.is_called_by_incoming_handler_thread():
                raise Exception('Can not get parameter from callback until fully connected.')
            if not self._initialized.wait(timeout=60):
                raise Exception('Connection timed out')

        [group, name] = complete_name.split('.')
        return self.values[group][name]

    def get_default_value(self, complete_name, callback):
        """
        Get the default value of the specified parameter.
        The supplied callback will be called with the name of the parameter
        as well as the default value. None if there is an error.

        @param complete_name The 'group.name' name of the parameter to store
        @param callback The callback should take `complete_name` and default value as argument
        """
        element = self.toc.get_element_by_complete_name(complete_name)

        def new_packet_cb(pk):
            if pk.channel == MISC_CHANNEL and pk.data[0] == MISC_GET_DEFAULT_VALUE:
                if pk.data[3] == errno.ENOENT:
                    callback(complete_name, None)
                    self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)
                    return

                default_value, = struct.unpack(element.pytype, pk.data[3:])
                callback(complete_name, default_value)
                self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)

        self.cf.add_port_callback(CRTPPort.PARAM, new_packet_cb)

        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)
        pk.data = struct.pack('<BH', MISC_GET_DEFAULT_VALUE, element.ident)
        self.param_updater.send_param_misc(pk)

    def persistent_clear(self, complete_name, callback=None):
        """
        Clear the current value of the specified persistent parameter from
        eeprom. The supplied callback will be called with `True` as an
        argument on success and with `False` as an argument on failure.

        @param complete_name The 'group.name' name of the parameter to store
        @param callback Optional callback should take `complete_name` and boolean status as arguments
        """
        element = self.toc.get_element_by_complete_name(complete_name)
        if not element.is_persistent():
            raise AttributeError(f"Param '{complete_name}' is not persistent")

        def new_packet_cb(pk):
            if pk.channel == MISC_CHANNEL and pk.data[0] == MISC_PERSISTENT_CLEAR:
                callback(complete_name, pk.data[3] == 0)
                self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)

        if callback is not None:
            self.cf.add_port_callback(CRTPPort.PARAM, new_packet_cb)

        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)
        pk.data = struct.pack('<BH', MISC_PERSISTENT_CLEAR, element.ident)
        self.param_updater.send_param_misc(pk)

    def persistent_store(self, complete_name, callback=None):
        """
        Store the current value of the specified persistent parameter to
        eeprom. The supplied callback will be called with `True` as an
        argument on success, and with `False` as an argument on failure.

        @param complete_name The 'group.name' name of the parameter to store
        @param callback Optional callback should take `complete_name` and boolean status as arguments
        """
        element = self.toc.get_element_by_complete_name(complete_name)
        if not element:
            callback(complete_name, False)
            return
        if not element.is_persistent():
            raise AttributeError(f"Param '{complete_name}' is not persistent")

        def new_packet_cb(pk):
            if pk.channel == MISC_CHANNEL and pk.data[0] == MISC_PERSISTENT_STORE:
                callback(complete_name, pk.data[3] == 0)
                self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)

        if callback is not None:
            self.cf.add_port_callback(CRTPPort.PARAM, new_packet_cb)

        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)
        pk.data = struct.pack('<BH', MISC_PERSISTENT_STORE, element.ident)
        self.param_updater.send_param_misc(pk)

    def persistent_get_state(self, complete_name, callback):
        """
        Get the state of the specified persistent parameter. The state will be
        returned in the supplied callback. The state is represented as a
        namedtuple with members: `is_stored`, `default_value` and
        `stored_value`. The state is `None` if the parameter is not persistent
        or if something goes wrong.

        | Member            | Description                                     |
        | ----------------- | ----------------------------------------------- |
        | `is_stored`       | `True` if the value is stored to eeprom         |
        | `default_value`   | The default value supplied by the firmware      |
        | `stored_value`    | Value stored in eeprom, None if `not is_stored` |

        @param complete_name The 'group.name' name of the parameter to store
        @param callback Callback, takes `complete_name` and PersistentParamState namedtuple as arg
        """
        element = self.toc.get_element_by_complete_name(complete_name)
        if not element.is_persistent():
            raise AttributeError(f"Param '{complete_name}' is not persistent")

        def new_packet_cb(pk):
            if pk.channel == MISC_CHANNEL and pk.data[0] == MISC_PERSISTENT_GET_STATE:
                if pk.data[3] == errno.ENOENT:
                    callback(complete_name, None)
                    self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)
                    return

                is_stored = pk.data[3] == 1
                if not is_stored:
                    default_value, = struct.unpack(element.pytype, pk.data[4:])
                else:
                    # Remove little-endian indicator ('<')
                    just_type = element.pytype[1:]
                    default_value, stored_value = struct.unpack(f'<{just_type * 2}', pk.data[4:])

                callback(complete_name,
                         PersistentParamState(
                             is_stored,
                             default_value,
                             stored_value if is_stored else None
                         )
                         )
                self.cf.remove_port_callback(CRTPPort.PARAM, new_packet_cb)

        self.cf.add_port_callback(CRTPPort.PARAM, new_packet_cb)
        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)
        pk.data = struct.pack('<BH', MISC_PERSISTENT_GET_STATE, element.ident)
        self.param_updater.send_param_misc(pk)


class _ExtendedTypeFetcher(Thread):

    def __init__(self, cf, toc):
        Thread.__init__(self, name='ExtendedTypeFetcherThread')
        self.daemon = True
        self._lock = Lock()

        self._cf = cf
        self._toc = toc
        self._done_callback = None

        self.request_queue = Queue()
        self._cf.add_port_callback(CRTPPort.PARAM, self._new_packet_cb)
        self._should_close = False
        self._req_param = -1
        self._count = -1

    def _new_packet_cb(self, pk):
        """Callback for newly arrived packets"""
        if pk.channel == MISC_CHANNEL:
            var_id = struct.unpack('<H', pk.data[1:3])[0]

            if self._req_param == var_id:
                extended_type = pk.data[3]
                if extended_type == ParamTocElement.EXTENDED_PERSISTENT:
                    self._toc.get_element_by_id(var_id).mark_persistent()
                self._count -= 1
                if self._count == 0:
                    if self._done_callback is not None:
                        self._done_callback()
                    self._close()

                self._req_param = -1
                try:
                    self._lock.release()
                except Exception:
                    pass

    def set_callback(self, callback):
        self._done_callback = callback

    def request_extended_types(self, elements):
        self._count = len(elements)
        for element in elements:
            pk = CRTPPacket()
            pk.set_header(CRTPPort.PARAM, MISC_CHANNEL)

            pk.data = struct.pack('<BH', MISC_GET_EXTENDED_TYPE, element.ident)
            self.request_queue.put(pk)

    def _close(self):
        # First empty the queue from all packets
        try:
            while True:
                self.request_queue.get(block=False)
        except Empty:
            pass
        self.request_queue.put(None)  # Make sure we exit the run loop
        self._should_close = True
        self._cf.remove_port_callback(CRTPPort.PARAM, self._new_packet_cb)

        # Then force an unlock of the mutex if we are waiting for a packet
        # we didn't get back due to a disconnect for example.
        try:
            self._lock.release()
        except RuntimeError:
            pass

    def run(self):
        while not self._should_close:
            pk = self.request_queue.get()  # Wait for request update
            if pk is None:
                continue
            self._lock.acquire()
            if self._cf.link:
                self._req_param = struct.unpack('<H', pk.data[1:3])[0]
                self._cf.send_packet(pk, expected_reply=(tuple(pk.data[:3])))
            else:
                self._lock.release()


class _ParamUpdater(Thread):
    """This thread will update params through a queue to make sure that we
    get back values"""

    def __init__(self, cf, useV2, updated_callback):
        """Initialize the thread"""
        Thread.__init__(self, name='ParamUpdaterThread')
        self.daemon = True
        self.wait_lock = Lock()
        self.cf = cf
        self._useV2 = useV2
        self.updated_callback = updated_callback
        self.request_queue = Queue()
        self.cf.add_port_callback(CRTPPort.PARAM, self._new_packet_cb)
        self._should_close = False
        self._lock_pattern = None

    def close(self):
        # First empty the queue from all packets
        try:
            while True:
                self.request_queue.get(block=False)
        except Empty:
            pass
        self.request_queue.put(None)  # Make sure we exit the run loop
        self._should_close = True
        self.cf.remove_port_callback(CRTPPort.PARAM, self._new_packet_cb)
        # Then force an unlock of the mutex if we are waiting for a packet
        # we didn't get back due to a disconnect for example.
        try:
            self.wait_lock.release()
        except RuntimeError:
            pass

    def request_param_setvalue(self, pk):
        """Place a param set value request on the queue. When this is sent to
        the Crazyflie it will answer with the update param value. """
        self.request_queue.put(pk)

    def send_param_misc(self, pk):
        """Place a param misc request on the queue. When this is sent to
        the Crazyflie it will answer with the same var_id and command. """
        self.request_queue.put(pk)

    def _new_packet_cb(self, pk):
        """Callback for newly arrived packets"""
        if pk.channel == READ_CHANNEL or pk.channel == WRITE_CHANNEL:
            if self._useV2:
                release_pattern = pk.data[:2]
                if pk.channel == READ_CHANNEL:
                    pk = copy.deepcopy(pk)  # Dont modify the original packet
                    pk.data = pk.data[:2] + pk.data[3:]
            else:
                release_pattern = pk.data[:1]

            if (pk.channel != TOC_CHANNEL and self._lock_pattern == release_pattern and
                    pk is not None):
                self.updated_callback(pk)
                self._lock_pattern = None
                try:
                    self.wait_lock.release()
                except Exception:
                    pass
        elif pk.channel == MISC_CHANNEL:
            command = pk.data[0]
            if command == MISC_VALUE_UPDATED:
                self.updated_callback(pk)

            release_pattern = pk.data[:3]
            if self._lock_pattern == release_pattern:
                self._lock_pattern = None
                self.wait_lock.release()

    def request_param_update(self, var_id):
        """Place a param update request on the queue"""
        self._useV2 = self.cf.platform.get_protocol_version() >= 4
        pk = CRTPPacket()
        pk.set_header(CRTPPort.PARAM, READ_CHANNEL)
        if self._useV2:
            pk.data = struct.pack('<H', var_id)
        else:
            pk.data = struct.pack('<B', var_id)
        logger.debug('Requesting request to update param [%d]', var_id)
        self.request_queue.put(pk)

    def run(self):
        while not self._should_close:
            pk = self.request_queue.get()  # Wait for request update
            if pk is None:
                continue
            self.wait_lock.acquire()
            if self.cf.link:
                if self._useV2:
                    if pk.channel == MISC_CHANNEL:
                        self._lock_pattern = pk.data[:3]
                    else:
                        self._lock_pattern = pk.data[:2]

                    self.cf.send_packet(pk, expected_reply=(tuple(self._lock_pattern)))
                else:
                    self._lock_pattern = pk.data[:1]
                    self.cf.send_packet(pk, expected_reply=(tuple(pk.data[:1])))
            else:
                self.wait_lock.release()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time
import unittest

import cflib.crtp
from cflib.crtp.crtpstack import CRTPPacket
from cflib.crtp.crtpstack import CRTPPort
from sys_test.swarm_test_rig.rig_support import RigSupport


class TestResponseTime(unittest.TestCase):
    ECHO = 0

    def setUp(self):
        cflib.crtp.init_drivers()
        self.test_rig_support = RigSupport()

        self.links = []

    def tearDown(self):
        for link in self.links:
            link.close()
        self.links = []

    def test_response_time_to_one_cf(self):
        # Fixture
        uri = self.test_rig_support.all_uris[0]
        self.test_rig_support.restart_devices([uri])
        link = self.connect_link(uri)
        seq_nr = 47
        expected_max_response_time = 0.01

        # Test
        time_send_echo = time.time()
        self.request_echo_with_seq_nr(link, seq_nr)
        response_timestamps = self.assert_wait_for_all_seq_nrs([link], seq_nr)
        response_time = response_timestamps[uri] - time_send_echo

        # Assert
        self.assertLess(response_time, expected_max_response_time)

    def test_response_time_to_all_cfs(self):
        # Fixture
        uris = self.test_rig_support.all_uris
        self.test_rig_support.restart_devices(uris)

        for uri in uris:
            self.connect_link(uri)

        seq_nr = 47
        expected_max_response_time = 0.1
        expected_mean_response_time = 0.05

        # Test
        time_send_echo = time.time()
        for link in self.links:
            self.request_echo_with_seq_nr(link, seq_nr)

        response_timestamps = self.assert_wait_for_all_seq_nrs(
            self.links, seq_nr)

        # Assert
        response_times = {}
        for uri, response_time in response_timestamps.items():
            response_times[uri] = response_time - time_send_echo

        times = response_times.values()
        max_time = max(times)
        mean_time = float(sum(times)) / len(times)

        # print(max_time, mean_time, times)
        self.assertLess(max_time, expected_max_response_time)
        self.assertLess(mean_time, expected_mean_response_time)

    def request_echo_with_seq_nr(self, link, seq_nr):
        pk = CRTPPacket()
        pk.set_header(CRTPPort.LINKCTRL, self.ECHO)
        pk.data = (seq_nr,)
        link.send_packet(pk)

    def assert_wait_for_all_seq_nrs(self, links, seq_nr, timeout=1):
        NO_BLOCKING = -1

        time_end = time.time() + timeout
        response_timestamps = {}
        while time.time() < time_end:
            for link in links:
                if link.uri not in response_timestamps:
                    response = link.receive_packet(wait=NO_BLOCKING)
                    if self._is_response_correct_seq_nr(response, seq_nr):
                        response_timestamps[link.uri] = time.time()

            if len(response_timestamps) == len(self.links):
                return response_timestamps

            time.sleep(0.001)

        self.fail('Time out while waiting for seq nrs.')

    def _is_response_correct_seq_nr(self, response, seq_nr):
        if response is not None:
            if response._get_channel() == self.ECHO and \
                    response._get_port() == CRTPPort.LINKCTRL:
                received_seq = response._get_data_t()[0]
                if received_seq == seq_nr:
                    return True

        return False

    def connect_link(self, uri):
        link = cflib.crtp.get_link_driver(uri, self._radio_link_statistics_cb,
                                          self._link_error_cb)
        self.assertIsNotNone(link)
        self.links.append(link)

        return link

    def _radio_link_statistics_cb(self, radoi_link_statistics):
        pass

    def _link_error_cb(self, errmsg):
        self.fail(errmsg)
